# Copyright 2015 Brocade Communications System, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import time
import yaml

from blazarclient import client as blazar_client
from cinderclient import client as cinder_client
from glanceclient.v2 import client as glance_client
from keystoneauth1.identity import v3
from keystoneauth1 import session
from keystoneclient import adapter
from neutronclient.v2_0 import client as neutron_client
from novaclient import client as nova_client
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from tempest.lib import base

from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import clients
from tacker.tests.utils import read_file
from tacker import version

from tackerclient.common import exceptions
from tackerclient.v1_0 import client as tacker_client


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class SessionClient(adapter.Adapter):
    def request(self, *args, **kwargs):
        kwargs.setdefault('authenticated', False)
        kwargs.setdefault('raise_exc', False)

        content_type = kwargs.pop('content_type', None) or 'application/json'

        headers = kwargs.setdefault('headers', {})
        headers.setdefault('Accept', content_type)

        try:
            kwargs.setdefault('data', kwargs.pop('body'))
        except KeyError:
            pass

        if kwargs.get('data'):
            headers.setdefault('Content-Type', content_type)

        return super(SessionClient, self).request(*args, **kwargs)

    def _decode_json(self, response):
        body = response.text
        if body and response.headers['Content-Type'] == 'text/plain':
            return body
        elif body and response.headers['Content-Type'] == 'text/x-sh':
            return body
        elif body:
            return jsonutils.loads(body)
        else:
            return ""

    def do_request(self, url, method, **kwargs):
        kwargs.setdefault('authenticated', True)
        resp = self.request(url, method, **kwargs)
        if resp.headers['Content-Type'] == 'application/zip':
            return resp, resp.content
        body = self._decode_json(resp)
        return resp, body


class BaseTackerTest(base.BaseTestCase):
    """Base test case class for all Tacker API tests."""

    @classmethod
    def setUpClass(cls):
        super(BaseTackerTest, cls).setUpClass()
        kwargs = {}

        cfg.CONF(args=['--config-file', '/etc/tacker/tacker.conf'],
                 project='tacker',
                 version='%%prog %s' % version.version_info.release_string(),
                 **kwargs)

        cls.client = cls.tackerclient()
        cls.http_client = cls.tacker_http_client()
        cls.h_client = cls.heatclient()
        cls.glance_client = cls.glanceclient()
        cls.cinder_client = cls.cinderclient()
        cls.nova_client = cls.novaclient()

    @classmethod
    def get_credentials(cls, vim_conf_file=None):
        if vim_conf_file is None:
            vim_conf_file = 'local-vim.yaml'
        vim_params = yaml.safe_load(read_file(vim_conf_file))
        vim_params['auth_url'] += '/v3'
        return vim_params

    @classmethod
    def get_auth_session(cls, vim_conf_file=None):
        vim_params = cls.get_credentials(vim_conf_file)
        auth = v3.Password(
            auth_url=vim_params['auth_url'],
            username=vim_params['username'],
            password=vim_params['password'],
            project_name=vim_params['project_name'],
            user_domain_name=vim_params['user_domain_name'],
            project_domain_name=vim_params['project_domain_name'])
        verify = 'True' == vim_params.pop('cert_verify', 'False')
        auth_ses = session.Session(auth=auth, verify=verify)
        return auth_ses

    @classmethod
    def tacker_http_client(cls, vim_conf_file=None):
        auth_session = cls.get_auth_session(vim_conf_file)
        return SessionClient(session=auth_session,
                             service_type='nfv-orchestration',
                             region_name='RegionOne')

    @classmethod
    def tackerclient(cls, vim_conf_file=None):
        auth_session = cls.get_auth_session(vim_conf_file)
        return tacker_client.Client(session=auth_session, retries=5)

    @classmethod
    def novaclient(cls, vim_conf_file=None):
        vim_params = cls.get_credentials(vim_conf_file)
        auth = v3.Password(auth_url=vim_params['auth_url'],
            username=vim_params['username'],
            password=vim_params['password'],
            project_name=vim_params['project_name'],
            user_domain_name=vim_params['user_domain_name'],
            project_domain_name=vim_params['project_domain_name'])
        verify = 'True' == vim_params.pop('cert_verify', 'False')
        auth_ses = session.Session(auth=auth, verify=verify)
        return nova_client.Client(constants.NOVA_CLIENT_VERSION,
                                  session=auth_ses)

    @classmethod
    def neutronclient(cls, vim_conf_file=None):
        vim_params = cls.get_credentials(vim_conf_file)
        auth = v3.Password(auth_url=vim_params['auth_url'],
            username=vim_params['username'],
            password=vim_params['password'],
            project_name=vim_params['project_name'],
            user_domain_name=vim_params['user_domain_name'],
            project_domain_name=vim_params['project_domain_name'])
        verify = 'True' == vim_params.pop('cert_verify', 'False')
        auth_ses = session.Session(auth=auth, verify=verify)
        return neutron_client.Client(session=auth_ses)

    @classmethod
    def heatclient(cls, vim_conf_file=None):
        if vim_conf_file is None:
            vim_conf_file = 'local-vim.yaml'
        data = yaml.safe_load(read_file(vim_conf_file))
        data['auth_url'] = data['auth_url'] + '/v3'
        domain_name = data.pop('domain_name')
        data['user_domain_name'] = domain_name
        data['project_domain_name'] = domain_name
        return clients.OpenstackClients(auth_attr=data).heat

    def blazarclient(cls, vim_conf_file=None):
        data = cls.get_credentials(vim_conf_file)
        domain_name = data.pop('domain_name')
        data['user_domain_name'] = domain_name
        data['project_domain_name'] = domain_name
        auth_ses = clients.OpenstackClients(auth_attr=data).keystone_session
        args = {
            'session': auth_ses,
            'service_type': 'reservation',
            'interface': 'public',
            'region_name': 'RegionOne',
        }
        client = blazar_client.Client(**args)
        return client

    @classmethod
    def glanceclient(cls, vim_conf_file=None):
        vim_params = cls.get_credentials(vim_conf_file)
        auth = v3.Password(auth_url=vim_params['auth_url'],
            username=vim_params['username'],
            password=vim_params['password'],
            project_name=vim_params['project_name'],
            user_domain_name=vim_params['user_domain_name'],
            project_domain_name=vim_params['project_domain_name'])
        verify = 'True' == vim_params.pop('cert_verify', 'False')
        auth_ses = session.Session(auth=auth, verify=verify)
        return glance_client.Client(session=auth_ses)

    @classmethod
    def aodh_http_client(cls, vim_conf_file=None):
        auth_session = cls.get_auth_session(vim_conf_file)
        return SessionClient(session=auth_session,
                             service_type='alarming',
                             region_name='RegionOne')

    @classmethod
    def cinderclient(cls, vim_conf_file=None):
        vim_params = cls.get_credentials(vim_conf_file)
        auth = v3.Password(auth_url=vim_params['auth_url'],
            username=vim_params['username'],
            password=vim_params['password'],
            project_name=vim_params['project_name'],
            user_domain_name=vim_params['user_domain_name'],
            project_domain_name=vim_params['project_domain_name'])
        verify = 'True' == vim_params.pop('cert_verify', 'False')
        auth_ses = session.Session(auth=auth, verify=verify)
        return cinder_client.Client(constants.CINDER_CLIENT_VERSION,
                                    session=auth_ses)

    def get_vdu_resource(self, stack_id, res_name):
        return self.h_client.resources.get(stack_id, res_name)

    def wait_until_vnf_status(self, vnf_id, target_status, timeout,
                              sleep_interval):
        start_time = int(time.time())
        status = None
        while True:
            try:
                vnf_result = self.client.show_vnf(vnf_id)
                status = vnf_result['vnf']['status']
                if status == target_status:
                    break
            except exceptions.InternalServerError:
                pass

            if int(time.time()) - start_time > timeout:
                break
            time.sleep(sleep_interval)

        self.assertEqual(status, target_status,
                         "vnf %(vnf_id)s with status %(status)s is"
                         " expected to be %(target)s" %
                         {"vnf_id": vnf_id, "status": status,
                          "target": target_status})

    def wait_until_vnf_active(self, vnf_id, timeout, sleep_interval):
        self.wait_until_vnf_status(vnf_id, 'ACTIVE', timeout,
                                   sleep_interval)

    def verify_vnf_update(self, vnf_id):
        self.wait_until_vnf_status(vnf_id, 'ACTIVE',
                                   constants.VNF_CIRROS_CREATE_TIMEOUT,
                                   constants.ACTIVE_SLEEP_TIME)
        self.wait_until_vnf_status(vnf_id, 'PENDING_HEAL',
                                   constants.VNF_CIRROS_PENDING_HEAL_TIMEOUT,
                                   constants.PENDING_SLEEP_TIME)
        self.wait_until_vnf_status(vnf_id, 'ACTIVE',
                                  constants.VNF_CIRROS_CREATE_TIMEOUT,
                                  constants.ACTIVE_SLEEP_TIME)

    def wait_until_vnf_delete(self, vnf_id, timeout, sleep_interval=1):
        start_time = int(time.time())
        while True:
            status = None
            try:
                vnf_result = self.client.show_vnf(vnf_id)
                status = vnf_result['vnf']['status']
            except exceptions.NotFound:
                return
            except Exception as e:
                LOG.error("Failed to get vnf status: %s", str(e))

            if status is not None and status != 'PENDING_DELETE':
                raise Exception("Failed with status: %s" % status)
            if int(time.time()) - start_time > timeout:
                raise Exception("Timeout for deleting vnf %s.",
                                vnf_id)
            time.sleep(sleep_interval)

    def wait_until_vnf_dead(self, vnf_id, timeout, sleep_interval):
        self.wait_until_vnf_status(vnf_id, 'DEAD', timeout,
                                   sleep_interval)

    def validate_vnf_instance(self, vnfd_instance, vnf_instance):
        self.assertIsNotNone(vnf_instance)
        self.assertIsNotNone(vnf_instance['vnf']['id'])
        self.assertIsNotNone(vnf_instance['vnf']['instance_id'])
        if vnfd_instance:
            self.assertEqual(vnf_instance['vnf']['vnfd_id'], vnfd_instance[
                'vnfd']['id'])

    def verify_vnf_restart(self, vnfd_instance, vnf_instance):
        vnf_id = vnf_instance['vnf']['id']
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)
        self.validate_vnf_instance(vnfd_instance, vnf_instance)
        self.assertIsNotNone(self.client.show_vnf(vnf_id)['vnf'][
            'mgmt_ip_address'])

        self.wait_until_vnf_dead(
            vnf_id,
            constants.VNF_CIRROS_DEAD_TIMEOUT,
            constants.DEAD_SLEEP_TIME)
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)
        self.validate_vnf_instance(vnfd_instance, vnf_instance)

    def verify_vnf_monitor_events(self, vnf_id, vnf_state_list):
        for state in vnf_state_list:
            params = {'resource_id': vnf_id, 'resource_state': state,
                      'event_type': evt_constants.RES_EVT_MONITOR}
            vnf_evt_list = self.client.list_vnf_events(**params)
            mesg = ("%s - state transition expected." % state)
            self.assertIsNotNone(vnf_evt_list['vnf_events'], mesg)

    def verify_vnf_crud_events(self, vnf_id, evt_type, res_state,
                               tstamp=None, cnt=1):
        params = {'resource_id': vnf_id,
                  'resource_state': res_state,
                  'resource_type': evt_constants.RES_TYPE_VNF,
                  'event_type': evt_type}
        if tstamp:
            params['timestamp'] = tstamp

        vnf_evt_list = self.client.list_vnf_events(**params)

        self.assertIsNotNone(vnf_evt_list['vnf_events'],
                             "List of VNF events are Empty")
        self.assertEqual(cnt, len(vnf_evt_list['vnf_events']))

    def verify_vnfd_events(self, vnfd_id, evt_type, res_state,
                           tstamp=None, cnt=1):
        params = {'resource_id': vnfd_id,
                  'resource_state': res_state,
                  'resource_type': evt_constants.RES_TYPE_VNFD,
                  'event_type': evt_type}
        if tstamp:
            params['timestamp'] = tstamp

        vnfd_evt_list = self.client.list_vnfd_events(**params)

        self.assertIsNotNone(vnfd_evt_list['vnfd_events'],
                             "List of VNFD events are Empty")
        self.assertEqual(cnt, len(vnfd_evt_list['vnfd_events']))

    def get_vim(self, vim_list, vim_name):
        if len(vim_list.values()) == 0:
            assert False, "vim_list is Empty: Default VIM is missing"

        for vim_list in vim_list.values():
            for vim in vim_list:
                if vim['name'] == vim_name:
                    return vim
        return None

    def verify_antispoofing_in_stack(self, stack_id, resource_name):
        resource_types = self.h_client.resources
        resource_details = resource_types.get(stack_id=stack_id,
                                              resource_name=resource_name)
        resource_dict = resource_details.to_dict()
        self.assertTrue(resource_dict['attributes']['port_security_enabled'])

    def trigger_vnf(self, vnf, policy_name, policy_action):
        credential = 'g0jtsxu9'
        body = {"trigger": {'policy_name': policy_name,
                            'action_name': policy_action,
                            'params': {
                                'data': {'alarm_id': '35a80852-e24f-46ed-bd34-e2f831d00172', 'current': 'alarm'},  # noqa
                                'credential': credential}
                            }
                }
        self.client.post('/vnfs/%s/triggers' % vnf, body)

    def assertDictSupersetOf(self, expected_subset, actual_superset):
        """Checks that actual dict contains the expected dict.

        After checking that the arguments are of the right type, this checks
        that each item in expected_subset is in, and matches, what is in
        actual_superset. Separate tests are done, so that detailed info can
        be reported upon failure.
        """
        if not isinstance(expected_subset, dict):
            self.fail("expected_subset (%s) is not an instance of dict" %
                      type(expected_subset))
        if not isinstance(actual_superset, dict):
            self.fail("actual_superset (%s) is not an instance of dict" %
                      type(actual_superset))
        for k, v in expected_subset.items():
            self.assertIn(k, actual_superset)
            self.assertEqual(v, actual_superset[k],
                             "Key %(key)s expected: %(exp)r, actual %(act)r" %
                             {'key': k, 'exp': v, 'act': actual_superset[k]})

    def create_cinder_volume(cls, vol_size, vol_name):
        try:
            cinder_volume = cls.cinder_client.volumes.create(vol_size,
                    name=vol_name)
        except Exception as e:
            LOG.error("Failed to create cinder volume: %s", str(e))
            return None

        return cinder_volume.id

    def delete_cinder_volume(cls, vol_id):
        try:
            cls.cinder_client.volumes.delete(vol_id)
        except Exception as e:
            LOG.error("Failed to delete cinder volume: %s", str(e))

    def vnfd_and_vnf_create(self, vnfd_file, vnf_name, volume_id=None,
            volume_name=None):
        input_yaml = read_file(vnfd_file)
        tosca_dict = yaml.safe_load(input_yaml)
        if volume_id is not None:
            volume_detail = tosca_dict['topology_template']['inputs']
            volume_detail[volume_name]['default'] = volume_id
        tosca_arg = {'vnfd': {'name': vnf_name,
                              'attributes': {'vnfd': tosca_dict}}}

        # Create vnfd with tosca template
        vnfd_instance = self.client.create_vnfd(body=tosca_arg)
        self.assertIsNotNone(vnfd_instance)

        # Create vnf with vnfd_id
        vnfd_id = vnfd_instance['vnfd']['id']
        self.addCleanup(self.client.delete_vnfd, vnfd_id)

        vnf_arg = {'vnf': {'vnfd_id': vnfd_id, 'name': vnf_name}}
        vnf_instance = self.client.create_vnf(body=vnf_arg)
        self.validate_vnf_instance(vnfd_instance, vnf_instance)

        return vnfd_instance, vnf_instance, tosca_dict

    def vnfd_and_vnf_create_inline(self, vnfd_file, vnf_name):
        vnfd_instance = {}
        input_yaml = read_file(vnfd_file)
        tosca_dict = yaml.safe_load(input_yaml)

        # create vnf directly from template
        vnf_arg = {'vnf': {'vnfd_template': tosca_dict, 'name': vnf_name}}
        vnf_instance = self.client.create_vnf(body=vnf_arg)
        self.validate_vnf_instance(vnfd_instance, vnf_instance)

        return vnf_instance, tosca_dict

    def _list_op_occs(self, filter_string='', http_client=None):
        if http_client is None:
            http_client = self.http_client
        show_url = os.path.join(
            self.base_vnf_lcm_op_occs_url)
        resp, response_body = http_client.do_request(
            show_url + filter_string, "GET")
        return resp, response_body

    def _assert_occ_list(self, resp, op_occs_list):
        self.assertEqual(200, resp.status_code)

        # Only check required parameters.
        for op_occs_info in op_occs_list:
            self.assertIsNotNone(op_occs_info.get('id'))
            self.assertIsNotNone(op_occs_info.get('operationState'))
            self.assertIsNotNone(op_occs_info.get('stateEnteredTime'))
            self.assertIsNotNone(op_occs_info.get('vnfInstanceId'))
            self.assertIsNotNone(op_occs_info.get('operation'))
            self.assertIsNotNone(op_occs_info.get('isAutomaticInvocation'))
            self.assertIsNotNone(op_occs_info.get('isCancelPending'))

            _links = op_occs_info.get('_links')
            self.assertIsNotNone(_links.get('self'))
            self.assertIsNotNone(_links.get('self').get('href'))
            self.assertIsNotNone(_links.get('vnfInstance'))
            self.assertIsNotNone(_links.get('vnfInstance').get('href'))
            self.assertIsNotNone(_links.get('grant'))
            self.assertIsNotNone(_links.get('grant').get('href'))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests
import time
import yaml

from oslo_config import cfg
from oslo_serialization import jsonutils

from tacker.common import clients
from tacker.common import log
from tacker.extensions import vnfm
from tacker.plugins.common import constants
from tacker.vnfm import vim_client


CONF = cfg.CONF
OPTS = [
    cfg.IntOpt('lead_time', default=120,
               help=_('Time for migration_type operation')),
    cfg.IntOpt('max_interruption_time', default=120,
               help=_('Time for how long live migration can take')),
    cfg.IntOpt('recovery_time', default=2,
               help=_('Time for migrated node could be fully running state')),
    cfg.IntOpt('request_retries',
               default=5,
               help=_("Number of attempts to retry for request")),
    cfg.IntOpt('request_retry_wait',
               default=5,
               help=_("Wait time (in seconds) between consecutive request"))
]
CONF.register_opts(OPTS, 'fenix')
MAINTENANCE_KEYS = (
    'instance_ids', 'session_id', 'state', 'reply_url'
)
MAINTENANCE_SUB_KEYS = {
    'PREPARE_MAINTENANCE': [('allowed_actions', 'list'),
                            ('instance_ids', 'list')],
    'PLANNED_MAINTENANCE': [('allowed_actions', 'list'),
                            ('instance_ids', 'list')]
}


def config_opts():
    return [('fenix', OPTS)]


class FenixPlugin(object):
    def __init__(self):
        self.REQUEST_RETRIES = cfg.CONF.fenix.request_retries
        self.REQUEST_RETRY_WAIT = cfg.CONF.fenix.request_retry_wait
        self.endpoint = None
        self._instances = {}
        self.vim_client = vim_client.VimClient()

    @log.log
    def request(self, plugin, context, vnf_dict, maintenance={},
                data_func=None):
        params_list = [maintenance]
        method = 'put'
        is_reply = True
        if data_func:
            action, create_func = data_func.split('_', 1)
            create_func = '_create_%s_list' % create_func
            if action in ['update', 'delete'] and hasattr(self, create_func):
                params_list = getattr(self, create_func)(
                    context, vnf_dict, action)
                method = action if action == 'delete' else 'put'
                is_reply = False
        for params in params_list:
            self._request(plugin, context, vnf_dict, params, method, is_reply)
        return len(params_list)

    @log.log
    def create_vnf_constraints(self, plugin, context, vnf_dict):
        self.update_vnf_constraints(plugin, context, vnf_dict,
                                    objects=['instance_group',
                                             'project_instance'])

    @log.log
    def delete_vnf_constraints(self, plugin, context, vnf_dict):
        self.update_vnf_constraints(plugin, context, vnf_dict,
                                    action='delete',
                                    objects=['instance_group',
                                             'project_instance'])

    @log.log
    def update_vnf_instances(self, plugin, context, vnf_dict,
                             action='update'):
        requests = self.update_vnf_constraints(plugin, context,
                                               vnf_dict, action,
                                               objects=['project_instance'])
        if requests[0]:
            self.post(context, vnf_dict)

    @log.log
    def update_vnf_constraints(self, plugin, context, vnf_dict,
                               action='update', objects=[]):
        result = []
        for obj in objects:
            requests = self.request(plugin, context, vnf_dict,
                                    data_func='%s_%s' % (action, obj))
            result.append(requests)
        return result

    @log.log
    def post(self, context, vnf_dict, **kwargs):
        post_function = getattr(context, 'maintenance_post_function', None)
        if not post_function:
            return
        post_function(context, vnf_dict)
        del context.maintenance_post_function

    @log.log
    def project_instance_pre(self, context, vnf_dict):
        key = vnf_dict['id']
        if key not in self._instances:
            self._instances.update({
                key: self._get_instances(context, vnf_dict)})

    @log.log
    def validate_maintenance(self, maintenance):
        body = maintenance['maintenance']['params']['data']['body']
        if not set(MAINTENANCE_KEYS).issubset(body) or \
                body['state'] not in constants.RES_EVT_MAINTENANCE:
            raise vnfm.InvalidMaintenanceParameter()
        sub_keys = MAINTENANCE_SUB_KEYS.get(body['state'], ())
        for key, val_type in sub_keys:
            if key not in body or type(body[key]) is not eval(val_type):
                raise vnfm.InvalidMaintenanceParameter()
        return body

    @log.log
    def _request(self, plugin, context, vnf_dict, maintenance,
                 method='put', is_reply=True):
        client = self._get_openstack_clients(context, vnf_dict)
        if not self.endpoint:
            self.endpoint = client.keystone_session.get_endpoint(
                service_type='maintenance', region_name=client.region_name)
            if not self.endpoint:
                raise vnfm.ServiceTypeNotFound(service_type_id='maintenance')

        if 'reply_url' in maintenance:
            url = maintenance['reply_url']
        elif 'url' in maintenance:
            url = "%s/%s" % (self.endpoint.rstrip('/'),
                             maintenance['url'].strip('/'))
        else:
            return

        def create_headers():
            return {
                'X-Auth-Token': client.keystone_session.get_token(),
                'Content-Type': 'application/json',
                'Accept': 'application/json'
            }

        request_body = {}
        request_body['headers'] = create_headers()
        state = constants.ACK if vnf_dict['status'] == constants.ACTIVE \
            else constants.NACK
        if method == 'put':
            data = maintenance.get('data', {})
            if is_reply:
                data['session_id'] = maintenance.get('session_id', '')
                data['state'] = "%s_%s" % (state, maintenance['state'])
            request_body['data'] = jsonutils.dump_as_bytes(data)

        def request_wait():
            retries = self.REQUEST_RETRIES
            while retries > 0:
                response = getattr(requests, method)(url, **request_body)
                if response.status_code == 200:
                    break
                else:
                    retries -= 1
                    time.sleep(self.REQUEST_RETRY_WAIT)

        plugin.spawn_n(request_wait)

    @log.log
    def handle_maintenance(self, plugin, context, maintenance):
        action = '_create_%s' % maintenance['state'].lower()
        maintenance['data'] = {}
        if hasattr(self, action):
            getattr(self, action)(plugin, context, maintenance)

    @log.log
    def _create_maintenance(self, plugin, context, maintenance):
        vnf_dict = maintenance.get('vnf', {})
        vnf_dict['attributes'].update({'maintenance_scaled': 0})
        plugin._update_vnf_post(context, vnf_dict['id'], constants.ACTIVE,
                                vnf_dict, constants.ACTIVE,
                                constants.RES_EVT_UPDATE)
        instances = self._get_instances(context, vnf_dict)
        instance_ids = [x['id'] for x in instances]
        maintenance['data'].update({'instance_ids': instance_ids})

    @log.log
    def _create_scale_in(self, plugin, context, maintenance):
        def post_function(context, vnf_dict):
            scaled = int(vnf_dict['attributes'].get('maintenance_scaled', 0))
            vnf_dict['attributes']['maintenance_scaled'] = str(scaled + 1)
            plugin._update_vnf_post(context, vnf_dict['id'], constants.ACTIVE,
                                    vnf_dict, constants.ACTIVE,
                                    constants.RES_EVT_UPDATE)
            instances = self._get_instances(context, vnf_dict)
            instance_ids = [x['id'] for x in instances]
            maintenance['data'].update({'instance_ids': instance_ids})
            self.request(plugin, context, vnf_dict, maintenance)

        vnf_dict = maintenance.get('vnf', {})
        policy_action = self._create_scale_dict(plugin, context, vnf_dict)
        if policy_action:
            maintenance.update({'policy_action': policy_action})
        context.maintenance_post_function = post_function

    @log.log
    def _create_prepare_maintenance(self, plugin, context, maintenance):
        self._create_planned_maintenance(plugin, context, maintenance)

    @log.log
    def _create_planned_maintenance(self, plugin, context, maintenance):
        def post_function(context, vnf_dict):
            migration_type = self._get_constraints(vnf_dict,
                                                   key='migration_type',
                                                   default='MIGRATE')
            maintenance['data'].update({'instance_action': migration_type})
            self.request(plugin, context, vnf_dict, maintenance)

        vnf_dict = maintenance.get('vnf', {})
        instances = self._get_instances(context, vnf_dict)
        request_instance_id = maintenance['instance_ids'][0]
        selected = None
        for instance in instances:
            if instance['id'] == request_instance_id:
                selected = instance
                break
        if not selected:
            vnfm.InvalidMaintenanceParameter()

        migration_type = self._get_constraints(vnf_dict, key='migration_type',
                                               default='MIGRATE')
        if migration_type == 'OWN_ACTION':
            policy_action = self._create_migrate_dict(context, vnf_dict,
                                                      selected)
            maintenance.update({'policy_action': policy_action})
            context.maintenance_post_function = post_function
        else:
            post_function(context, vnf_dict)

    @log.log
    def _create_maintenance_complete(self, plugin, context, maintenance):
        def post_function(context, vnf_dict):
            vim_res = self.vim_client.get_vim(context, vnf_dict['vim_id'])
            scaled = int(vnf_dict['attributes'].get('maintenance_scaled', 0))
            if vim_res['vim_type'] == 'openstack':
                scaled -= 1
                vnf_dict['attributes']['maintenance_scaled'] = str(scaled)
                plugin._update_vnf_post(context, vnf_dict['id'],
                                        constants.ACTIVE, vnf_dict,
                                        constants.ACTIVE,
                                        constants.RES_EVT_UPDATE)
            if scaled > 0:
                scale_out(plugin, context, vnf_dict)
            else:
                instances = self._get_instances(context, vnf_dict)
                instance_ids = [x['id'] for x in instances]
                maintenance['data'].update({'instance_ids': instance_ids})
                self.request(plugin, context, vnf_dict, maintenance)

        def scale_out(plugin, context, vnf_dict):
            policy_action = self._create_scale_dict(plugin, context, vnf_dict,
                                                    scale_type='out')
            context.maintenance_post_function = post_function
            plugin._vnf_action.invoke(policy_action['action'],
                                      'execute_action', plugin=plugin,
                                      context=context, vnf_dict=vnf_dict,
                                      args=policy_action['args'])

        vnf_dict = maintenance.get('vnf', {})
        scaled = vnf_dict.get('attributes', {}).get('maintenance_scaled', 0)
        if int(scaled):
            policy_action = self._create_scale_dict(plugin, context, vnf_dict,
                                                    scale_type='out')
            maintenance.update({'policy_action': policy_action})
            context.maintenance_post_function = post_function

    @log.log
    def _create_scale_dict(self, plugin, context, vnf_dict, scale_type='in'):
        policy_action, scale_dict = {}, {}
        policies = self._get_scaling_policies(plugin, context, vnf_dict)
        if not policies:
            return
        scale_dict['type'] = scale_type
        scale_dict['policy'] = policies[0]['name']
        policy_action['action'] = 'autoscaling'
        policy_action['args'] = {'scale': scale_dict}
        return policy_action

    @log.log
    def _create_migrate_dict(self, context, vnf_dict, instance):
        policy_action, heal_dict = {}, {}
        heal_dict['vdu_name'] = instance['name']
        heal_dict['cause'] = ["Migrate resource '%s' to other host."]
        heal_dict['stack_id'] = instance['stack_name']
        if 'scaling_group_names' in vnf_dict['attributes']:
            sg_names = vnf_dict['attributes']['scaling_group_names']
            sg_names = list(jsonutils.loads(sg_names).keys())
            heal_dict['heat_tpl'] = '%s_res.yaml' % sg_names[0]
        policy_action['action'] = 'vdu_autoheal'
        policy_action['args'] = heal_dict
        return policy_action

    @log.log
    def _create_instance_group_list(self, context, vnf_dict, action):
        group_id = vnf_dict['attributes'].get('maintenance_group', '')
        if not group_id:
            return

        def get_constraints(data):
            maintenance_config = self._get_constraints(vnf_dict)
            data['max_impacted_members'] = maintenance_config.get(
                'max_impacted_members', 1)
            data['recovery_time'] = maintenance_config.get('recovery_time', 60)

        params, data = {}, {}
        params['url'] = '/instance_group/%s' % group_id
        if action == 'update':
            data['group_id'] = group_id
            data['project_id'] = vnf_dict['tenant_id']
            data['group_name'] = 'tacker_nonha_app_group_%s' % vnf_dict['id']
            data['anti_affinity_group'] = False
            data['max_instances_per_host'] = 0
            data['resource_mitigation'] = True
            get_constraints(data)
        params.update({'data': data})
        return [params]

    @log.log
    def _create_project_instance_list(self, context, vnf_dict, action):
        group_id = vnf_dict.get('attributes', {}).get('maintenance_group', '')
        if not group_id:
            return

        params_list = []
        url = '/instance'
        instances = self._get_instances(context, vnf_dict)
        _instances = self._instances.get(vnf_dict['id'], {})
        if _instances:
            if action == 'update':
                instances = [v for v in instances if v not in _instances]
                del self._instances[vnf_dict['id']]
            else:
                instances = [v for v in _instances if v not in instances]
                if len(instances) != len(_instances):
                    del self._instances[vnf_dict['id']]

        if action == 'update':
            maintenance_configs = self._get_constraints(vnf_dict)
            for instance in instances:
                params, data = {}, {}
                params['url'] = '%s/%s' % (url, instance['id'])
                data['project_id'] = instance['project_id']
                data['instance_id'] = instance['id']
                data['instance_name'] = instance['name']
                data['migration_type'] = maintenance_configs.get(
                    'migration_type', 'MIGRATE')
                data['resource_mitigation'] = maintenance_configs.get(
                    'mitigation_type', True)
                data['max_interruption_time'] = maintenance_configs.get(
                    'max_interruption_time',
                    cfg.CONF.fenix.max_interruption_time)
                data['lead_time'] = maintenance_configs.get(
                    'lead_time', cfg.CONF.fenix.lead_time)
                data['group_id'] = group_id
                params.update({'data': data})
                params_list.append(params)
        elif action == 'delete':
            for instance in instances:
                params = {}
                params['url'] = '%s/%s' % (url, instance['id'])
                params_list.append(params)
        return params_list

    @log.log
    def _get_instances(self, context, vnf_dict):
        vim_res = self.vim_client.get_vim(context, vnf_dict['vim_id'])
        action = '_get_instances_with_%s' % vim_res['vim_type']
        if hasattr(self, action):
            return getattr(self, action)(context, vnf_dict)
        return {}

    @log.log
    def _get_instances_with_openstack(self, context, vnf_dict):
        def get_attrs_with_link(links):
            attrs = {}
            for link in links:
                href, rel = link['href'], link['rel']
                if rel == 'self':
                    words = href.split('/')
                    attrs['project_id'] = words[5]
                    attrs['stack_name'] = words[7]
                    break
            return attrs

        instances = []
        if not vnf_dict['instance_id']:
            return instances

        client = self._get_openstack_clients(context, vnf_dict)
        resources = client.heat.resources.list(vnf_dict['instance_id'],
                                               nested_depth=2)
        for resource in resources:
            if resource.resource_type == 'OS::Nova::Server' and \
               resource.resource_status != 'DELETE_IN_PROGRESS':
                instance = {
                    'id': resource.physical_resource_id,
                    'name': resource.resource_name
                }
                instance.update(get_attrs_with_link(resource.links))
                instances.append(instance)
        return instances

    @log.log
    def _get_scaling_policies(self, plugin, context, vnf_dict):
        vnf_id = vnf_dict['id']
        policies = []
        if 'scaling_group_names' in vnf_dict['attributes']:
            policies = plugin.get_vnf_policies(
                context, vnf_id, filters={'type': constants.POLICY_SCALING})
        return policies

    @log.log
    def _get_constraints(self, vnf, key=None, default=None):
        config = vnf.get('attributes', {}).get('config', '{}')
        maintenance_config = yaml.safe_load(config).get('maintenance', {})
        if key:
            return maintenance_config.get(key, default)
        return maintenance_config

    @log.log
    def _get_openstack_clients(self, context, vnf_dict):
        vim_res = self.vim_client.get_vim(context, vnf_dict['vim_id'])
        region_name = vnf_dict.setdefault('placement_attr', {}).get(
            'region_name', None)
        client = clients.OpenstackClients(auth_attr=vim_res['vim_auth'],
                                          region_name=region_name)
        return client



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import re

from tacker.sol_refactored.common import exceptions as sol_ex


supported_versions_v1 = {
    'uriPrefix': '/vnflcm/v1',
    'apiVersions': [
        {'version': '1.3.0', 'isDeprecated': False}
    ]
}

supported_versions_v2 = {
    'uriPrefix': '/vnflcm/v2',
    'apiVersions': [
        {'version': '2.0.0', 'isDeprecated': False}
    ]
}

CURRENT_VERSION = '2.0.0'

supported_versions = [
    item['version'] for item in supported_versions_v2['apiVersions']
]


class APIVersion(object):

    def __init__(self, version_string=None):
        self.ver_major = 0
        self.ver_minor = 0
        self.ver_patch = 0

        if version_string is None:
            return

        version_string = self._get_version_id(version_string)
        match = re.match(r"^([1-9]\d*)\.([1-9]\d*|0)\.([1-9]\d*|0)$",
                         version_string)
        if match:
            self.ver_major = int(match.group(1))
            self.ver_minor = int(match.group(2))
            self.ver_patch = int(match.group(3))
        else:
            raise sol_ex.InvalidAPIVersionString(version=version_string)

        if version_string not in supported_versions:
            raise sol_ex.APIVersionNotSupported(version=version_string)

    def _get_version_id(self, version_string):
        # version example (see. SOL013 Table 4.2.2-1)
        # `1.2.0` or `1.2.0-impl:example.com:myProduct:4`
        # This method checks the later case and return the part of
        # version identifier. check is loose.
        if '-' not in version_string:
            return version_string
        items = version_string.split('-')
        if len(items) == 2 and items[1].startswith("impl:"):
            return items[0]
        raise sol_ex.InvalidAPIVersionString(version=version_string)

    def is_null(self):
        return (self.ver_major, self.ver_minor, self.ver_patch) == (0, 0, 0)

    def __str__(self):
        return "%d.%d.%d" % (self.ver_major, self.ver_minor, self.ver_patch)

    def __lt__(self, other):
        return ((self.ver_major, self.ver_minor, self.ver_patch) <
                (other.ver_major, other.ver_minor, other.ver_patch))

    def __eq__(self, other):
        return ((self.ver_major, self.ver_minor, self.ver_patch) ==
                (other.ver_major, other.ver_minor, other.ver_patch))

    def __gt__(self, other):
        return ((self.ver_major, self.ver_minor, self.ver_patch) >
                (other.ver_major, other.ver_minor, other.ver_patch))

    def __le__(self, other):
        return self < other or self == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __ge__(self, other):
        return self > other or self == other

    def matches(self, min_version, max_version):
        if self.is_null():
            return False
        if max_version.is_null() and min_version.is_null():
            return True
        elif max_version.is_null():
            return min_version <= self
        elif min_version.is_null():
            return self <= max_version
        else:
            return min_version <= self <= max_version



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.sol_refactored.objects import base
from tacker.sol_refactored.objects import fields
from tacker.sol_refactored.objects.v2 import fields as v2fields


# NFV-SOL 003
# - v3.3.1 5.5.2.2 (API version: 2.0.0)
@base.TackerObjectRegistry.register
class VnfInstanceV2(base.TackerPersistentObject,
                    base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vnfInstanceName': fields.StringField(nullable=True),
        'vnfInstanceDescription': fields.StringField(nullable=True),
        'vnfdId': fields.StringField(nullable=False),
        'vnfProvider': fields.StringField(nullable=False),
        'vnfProductName': fields.StringField(nullable=False),
        'vnfSoftwareVersion': fields.VersionField(nullable=False),
        'vnfdVersion': fields.VersionField(nullable=False),
        'vnfConfigurableProperties': fields.KeyValuePairsField(nullable=True),
        'vimConnectionInfo': fields.DictOfObjectsField(
            'VimConnectionInfo', nullable=True),
        'instantiationState': fields.EnumField(
            valid_values=[
                'NOT_INSTANTIATED',
                'INSTANTIATED',
            ],
            nullable=False),
        'instantiatedVnfInfo': fields.ObjectField(
            'VnfInstanceV2_InstantiatedVnfInfo', nullable=True),
        'metadata': fields.KeyValuePairsField(nullable=True),
        'extensions': fields.KeyValuePairsField(nullable=True),
        '_links': fields.ObjectField('VnfInstanceV2_Links', nullable=False),
    }


@base.TackerObjectRegistry.register
class VnfInstanceV2_InstantiatedVnfInfo(base.TackerObject,
                                        base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'flavourId': fields.StringField(nullable=False),
        'vnfState': v2fields.VnfOperationalStateTypeField(nullable=False),
        'scaleStatus': fields.ListOfObjectsField(
            'ScaleInfoV2', nullable=True),
        'maxScaleLevels': fields.ListOfObjectsField(
            'ScaleInfoV2', nullable=True),
        'extCpInfo': fields.ListOfObjectsField(
            'VnfExtCpInfoV2', nullable=False),
        'extVirtualLinkInfo': fields.ListOfObjectsField(
            'ExtVirtualLinkInfoV2', nullable=True),
        'extManagedVirtualLinkInfo': fields.ListOfObjectsField(
            'ExtManagedVirtualLinkInfoV2', nullable=True),
        'monitoringParameters': fields.ListOfObjectsField(
            'MonitoringParameterV2', nullable=True),
        'localizationLanguage': fields.StringField(nullable=True),
        'vnfcResourceInfo': fields.ListOfObjectsField(
            'VnfcResourceInfoV2', nullable=True),
        'vnfVirtualLinkResourceInfo': fields.ListOfObjectsField(
            'VnfVirtualLinkResourceInfoV2', nullable=True),
        'virtualStorageResourceInfo': fields.ListOfObjectsField(
            'VirtualStorageResourceInfoV2', nullable=True),
        # NOTE: vnfcInfo exists in SOL002 only.
        'vnfcInfo': fields.ListOfObjectsField('VnfcInfoV2', nullable=True),
    }


@base.TackerObjectRegistry.register
class VnfInstanceV2_Links(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'self': fields.ObjectField('Link', nullable=False),
        'indicators': fields.ObjectField('Link', nullable=True),
        'instantiate': fields.ObjectField('Link', nullable=True),
        'terminate': fields.ObjectField('Link', nullable=True),
        'scale': fields.ObjectField('Link', nullable=True),
        'scaleToLevel': fields.ObjectField('Link', nullable=True),
        'changeFlavour': fields.ObjectField('Link', nullable=True),
        'heal': fields.ObjectField('Link', nullable=True),
        'operate': fields.ObjectField('Link', nullable=True),
        'changeExtConn': fields.ObjectField('Link', nullable=True),
        'createSnapshot': fields.ObjectField('Link', nullable=True),
        'revertToSnapshot': fields.ObjectField('Link', nullable=True),
    }



# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy.orm import exc as orm_exc

from oslo_db import exception
from oslo_log import log as logging

from tacker.common import log
from tacker.db.common_services import common_services_db
from tacker.db import db_base
from tacker.extensions import common_services
from tacker import manager


LOG = logging.getLogger(__name__)

EVENT_ATTRIBUTES = ('id', 'resource_id', 'resource_type', 'resource_state',
                    'timestamp', 'event_type', 'event_details')


class CommonServicesPluginDb(common_services.CommonServicesPluginBase,
                             db_base.CommonDbMixin):

    def __init__(self):
        super(CommonServicesPluginDb, self).__init__()

    @property
    def _core_plugin(self):
        return manager.TackerManager.get_plugin()

    def _make_event_dict(self, event_db, fields=None):
        res = dict((key, event_db[key]) for key in EVENT_ATTRIBUTES)
        return self._fields(res, fields)

    def _fields(self, resource, fields):
        if fields:
            return dict(((key, item) for key, item in resource.items()
                         if key in fields))
        return resource

    @log.log
    def create_event(self, context, res_id, res_type, res_state, evt_type,
                     tstamp, details=""):
        try:
            with context.session.begin(subtransactions=True):
                event_db = common_services_db.Event(
                    resource_id=res_id,
                    resource_type=res_type,
                    resource_state=res_state,
                    event_details=details,
                    event_type=evt_type,
                    timestamp=tstamp)
                context.session.add(event_db)
        except Exception as e:
            LOG.exception("create event error: %s", str(e))
            raise common_services.EventCreationFailureException(
                error_str=str(e))
        return self._make_event_dict(event_db)

    @log.log
    def get_event(self, context, event_id, fields=None):
        try:
            events_db = self._get_by_id(context,
                                        common_services_db.Event, event_id)
        except orm_exc.NoResultFound:
            raise common_services.EventNotFoundException(evt_id=event_id)
        return self._make_event_dict(events_db, fields)

    @log.log
    def get_events(self, context, filters=None, fields=None, sorts=None,
                   limit=None, marker_obj=None, page_reverse=False):
        try:
            return self._get_collection(context, common_services_db.Event,
                                        self._make_event_dict,
                                        filters, fields, sorts, limit,
                                        marker_obj, page_reverse)
        except exception.DBError as e:
            LOG.error("Failed to get event: %s", e.message)
            msg = "Id should be in UUID v4 format"
            raise common_services.InvalidFormat(error=msg)



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""

from oslo_log import log as logging

from tacker.common import exceptions

LOG = logging.getLogger(__name__)

supported_k8s_resource_kinds = {
    "Pod",
    "Binding",
    "ConfigMap",
    "LimitRange",
    "PersistentVolumeClaim",
    "PodTemplate",
    "ResourceQuota",
    "Secret",
    "ServiceAccount",
    "Service",
    "ControllerRevision",
    "DaemonSet",
    "Deployment",
    "ReplicaSet",
    "StatefulSet",
    "LocalSubjectAccessReview",
    "HorizontalPodAutoscaler",
    "Job",
    "Lease",
    "NetworkPolicy",
    "RoleBinding",
    "Role"
}


def check_and_save_namespace(
        instantiate_vnf_req, chk_namespaces, vnf_instance):
    namespace = ''
    if instantiate_vnf_req.additional_params:
        namespace = instantiate_vnf_req.additional_params.get('namespace', '')
    if not namespace:
        try:
            namespace = get_namespace_from_manifests(chk_namespaces)
        except exceptions.NamespaceIsNotUnique as e:
            LOG.error(e)
            raise e
    if not namespace:
        namespace = 'default'

    if not vnf_instance.vnf_metadata:
        vnf_instance.vnf_metadata = {}
    vnf_instance.vnf_metadata['namespace'] = namespace
    vnf_instance.save()


def get_namespace_from_manifests(chk_namespaces):
    namespaces = {
        chk_namespace['namespace'] for chk_namespace in
        chk_namespaces if (chk_namespace['kind'] in
                           supported_k8s_resource_kinds)
    }

    if len(namespaces) > 1:
        LOG.error(f'Multiple namespaces found: {namespaces}')
        raise exceptions.NamespaceIsNotUnique()

    if namespaces:
        return namespaces.pop()
    return None



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
from io import BytesIO
import json
import mimetypes
import os
import webob
import zipfile
from zipfile import ZipFile

from glance_store import exceptions as store_exceptions
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import excutils
from oslo_utils import uuidutils

from tacker._i18n import _
from tacker.api.schemas import vnf_packages
from tacker.api import validation
from tacker.api.views import vnf_packages as vnf_packages_view
from tacker.common import csar_utils
from tacker.common import exceptions
from tacker.common import utils
from tacker.conductor.conductorrpc import vnf_pkgm_rpc
from tacker.glance_store import store as glance_store
from tacker.objects import fields
from tacker.objects import vnf_package as vnf_package_obj
from tacker.policies import vnf_package as vnf_package_policies
from tacker import wsgi

LOG = logging.getLogger(__name__)


CONF = cfg.CONF


class VnfPkgmController(wsgi.Controller):

    _view_builder_class = vnf_packages_view.ViewBuilder

    def __init__(self):
        super(VnfPkgmController, self).__init__()
        self.rpc_api = vnf_pkgm_rpc.VNFPackageRPCAPI()
        glance_store.initialize_glance_store()

    def _get_vnf_package(self, id, request):
        # check if id is of type uuid format
        if not uuidutils.is_uuid_like(id):
            msg = _("Can not find requested vnf package: %s") % id
            raise webob.exc.HTTPNotFound(explanation=msg)

        try:
            vnf_package = vnf_package_obj.VnfPackage.get_by_id(
                request.context, id)
        except exceptions.VnfPackageNotFound:
            msg = _("Can not find requested vnf package: %s") % id
            raise webob.exc.HTTPNotFound(explanation=msg)
        return vnf_package

    @wsgi.response(http_client.CREATED)
    @wsgi.expected_errors((http_client.BAD_REQUEST, http_client.FORBIDDEN))
    @validation.schema(vnf_packages.create)
    def create(self, request, body):
        context = request.environ['tacker.context']
        context.can(vnf_package_policies.VNFPKGM % 'create')

        vnf_package = vnf_package_obj.VnfPackage(context=request.context)
        vnf_package.onboarding_state = (
            fields.PackageOnboardingStateType.CREATED)
        vnf_package.operational_state = (
            fields.PackageOperationalStateType.DISABLED)
        vnf_package.usage_state = fields.PackageUsageStateType.NOT_IN_USE
        vnf_package.user_data = body.get('userDefinedData', dict())
        vnf_package.tenant_id = request.context.project_id

        vnf_package.create()

        headers = {"location": '/vnfpkgm/v1/vnf_packages/%s' % vnf_package.id}
        result = self._view_builder.create(vnf_package)

        return wsgi.ResponseObject(result, headers=headers)

    @wsgi.response(http_client.OK)
    @wsgi.expected_errors((http_client.FORBIDDEN, http_client.NOT_FOUND))
    def show(self, request, id):
        context = request.environ['tacker.context']
        context.can(vnf_package_policies.VNFPKGM % 'show')

        # check if id is of type uuid format
        if not uuidutils.is_uuid_like(id):
            msg = _("Can not find requested vnf package: %s") % id
            raise webob.exc.HTTPNotFound(explanation=msg)

        try:
            vnf_package = vnf_package_obj.VnfPackage.get_by_id(
                request.context, id, expected_attrs=[
                    "vnf_deployment_flavours", "vnfd", "vnf_artifacts"])
        except exceptions.VnfPackageNotFound:
            msg = _("Can not find requested vnf package: %s") % id
            raise webob.exc.HTTPNotFound(explanation=msg)

        return self._view_builder.show(vnf_package)

    @wsgi.response(http_client.OK)
    @wsgi.expected_errors((http_client.BAD_REQUEST, http_client.FORBIDDEN))
    @validation.query_schema(vnf_packages.query_params_v1)
    def index(self, request):
        context = request.environ['tacker.context']
        context.can(vnf_package_policies.VNFPKGM % 'index')

        search_opts = {}
        search_opts.update(request.GET)

        def _key_exists(key, validate_value=True):
            try:
                request.GET[key]
            except KeyError:
                return False

            return True

        all_fields = _key_exists('all_fields')
        exclude_default = _key_exists('exclude_default')
        fields = request.GET.get('fields')
        exclude_fields = request.GET.get('exclude_fields')
        filters = request.GET.get('filter')
        if not (all_fields or fields or exclude_fields):
            exclude_default = True

        self._view_builder.validate_attribute_fields(all_fields=all_fields,
            fields=fields, exclude_fields=exclude_fields,
            exclude_default=exclude_default)

        filters = self._view_builder.validate_filter(filters)

        vnf_packages = vnf_package_obj.VnfPackagesList.get_by_filters(
            request.context, read_deleted='no', filters=filters)

        return self._view_builder.index(vnf_packages,
                                        all_fields=all_fields,
                                        exclude_fields=exclude_fields,
                                        fields=fields,
                                        exclude_default=exclude_default)

    @wsgi.response(http_client.NO_CONTENT)
    @wsgi.expected_errors((http_client.FORBIDDEN, http_client.NOT_FOUND,
                           http_client.CONFLICT))
    def delete(self, request, id):
        context = request.environ['tacker.context']
        context.can(vnf_package_policies.VNFPKGM % 'delete')

        vnf_package = self._get_vnf_package(id, request)

        if (vnf_package.operational_state ==
                fields.PackageOperationalStateType.ENABLED or
                vnf_package.usage_state ==
                fields.PackageUsageStateType.IN_USE):
            msg = _("VNF Package %(id)s cannot be deleted as it's "
                    "operational state is %(operational_state)s and usage "
                    "state is %(usage_state)s.")
            raise webob.exc.HTTPConflict(
                explanation=msg % {
                    "id": id,
                    "operational_state": vnf_package.operational_state,
                    "usage_state": vnf_package.usage_state})

        # Delete vnf_package
        self.rpc_api.delete_vnf_package(context, vnf_package)

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.expected_errors((http_client.FORBIDDEN, http_client.NOT_FOUND,
                           http_client.CONFLICT,
                           http_client.REQUESTED_RANGE_NOT_SATISFIABLE))
    def fetch_vnf_package_content(self, request, id):
        context = request.environ['tacker.context']
        context.can(vnf_package_policies.VNFPKGM % 'fetch_package_content')

        vnf_package = self._get_vnf_package(id, request)

        if vnf_package.onboarding_state != \
           fields.PackageOnboardingStateType.ONBOARDED:
            msg = _("VNF Package %(id)s onboarding state "
                    "is not %(onboarding)s")
            raise webob.exc.HTTPConflict(explanation=msg % {"id": id,
                    "onboarding": fields.PackageOnboardingStateType.ONBOARDED})

        if vnf_package.size == 0:

            try:
                zip_file_size = glance_store.get_csar_size(id,
                    vnf_package.location_glance_store)
                vnf_package.size = zip_file_size
                vnf_package.save()
            except exceptions.VnfPackageLocationInvalid:
                msg = _("Vnf package not present at location")
                raise webob.exc.HTTPNotFound(explanation=msg)

        else:
            zip_file_size = vnf_package.size

        range_val = self._get_range_from_request(request, zip_file_size)

        return self._download(
            request.response, range_val, id, vnf_package.location_glance_store,
            zip_file_size)

    def _download(self, response, range_val, uuid, location, zip_file_size):
        offset, chunk_size = 0, None
        if range_val:
            if isinstance(range_val, webob.byterange.Range):
                response_end = zip_file_size - 1
                # NOTE(sameert): webob parsing is zero-indexed.
                # i.e.,to download first 5 bytes of a 10 byte image,
                # request should be "bytes=0-4" and the response would be
                # "bytes 0-4/10".
                # Range if validated, will never have 'start' object as None.
                if range_val.start >= 0:
                    offset = range_val.start
                else:
                    # NOTE(sameert): Negative start values needs to be
                    # processed to allow suffix-length for Range request
                    # like "bytes=-2" as per rfc7233.
                    if abs(range_val.start) < zip_file_size:
                        offset = zip_file_size + range_val.start

                if range_val.end is not None and range_val.end < zip_file_size:
                    chunk_size = range_val.end - offset
                    response_end = range_val.end - 1
                else:
                    chunk_size = zip_file_size - offset

            response.status_int = 206

        response.headers['Content-Type'] = 'application/zip'

        response.app_iter = self._get_csar_zip_data(uuid,
            location, offset, chunk_size)
        # NOTE(sameert): In case of a full zip download, when
        # chunk_size was none, reset it to zip.size to set the
        # response header's Content-Length.
        if chunk_size is not None:
            response.headers['Content-Range'] = 'bytes %s-%s/%s'\
                                                % (offset,
                                                    response_end,
                                                    zip_file_size)
        else:
            chunk_size = zip_file_size
        response.headers['Content-Length'] = str(chunk_size)
        return response

    def _get_csar_zip_data(self, uuid, location, offset=0, chunk_size=None):
        try:
            resp, size = glance_store.load_csar_iter(
                uuid, location, offset=offset, chunk_size=chunk_size)
        except exceptions.VnfPackageLocationInvalid:
            msg = _("Vnf package not present at location")
            raise webob.exc.HTTPServerError(explanation=msg)
        return resp

    def _get_range_from_request(self, request, zip_file_size):
        range_str = request._headers.environ.get('HTTP_RANGE')
        if range_str is not None:
            # NOTE(sameert): We do not support multi range requests.
            if ',' in range_str:
                msg = _("Requests with multiple ranges are not supported in "
                       "Tacker. You may make multiple single-range requests "
                       "instead.")
                raise webob.exc.HTTPBadRequest(explanation=msg)

            range_ = webob.byterange.Range.parse(range_str)
            if range_ is None:
                range_err_msg = _("The byte range passed in the 'Range' header"
                 " did not match any available byte range in the VNF package"
                 " file")
                raise webob.exc.HTTPRequestRangeNotSatisfiable(
                    explanation=range_err_msg)
            # NOTE(sameert): Ensure that a range like bytes=4- for an zip
            # size of 3 is invalidated as per rfc7233.
            if range_.start >= zip_file_size:
                msg = _("Invalid start position in Range header. "
                       "Start position MUST be in the inclusive range"
                       "[0, %s].") % (zip_file_size - 1)
                raise webob.exc.HTTPRequestRangeNotSatisfiable(
                    explanation=msg)
            return range_

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.expected_errors((http_client.FORBIDDEN, http_client.NOT_FOUND,
                           http_client.CONFLICT))
    def upload_vnf_package_content(self, request, id, body):
        context = request.environ['tacker.context']
        context.can(vnf_package_policies.VNFPKGM % 'upload_package_content')

        # check if id is of type uuid format
        if not uuidutils.is_uuid_like(id):
            msg = _("Can not find requested vnf package: %s") % id
            return self._make_problem_detail('Not Found', msg, 404)

        try:
            vnf_package = vnf_package_obj.VnfPackage.get_by_id(
                request.context, id)
        except exceptions.VnfPackageNotFound:
            msg = _("Can not find requested vnf package: %s") % id
            return self._make_problem_detail('Not Found', msg, 404)
        except Exception as e:
            return self._make_problem_detail(
                'Internal Server Error', str(e), 500)

        if vnf_package.onboarding_state != \
           fields.PackageOnboardingStateType.CREATED:
            msg = _("VNF Package %(id)s onboarding state "
                    "is not %(onboarding)s")
            return self._make_problem_detail('Conflict', msg % {"id": id,
                    "onboarding": fields.PackageOnboardingStateType.CREATED},
                    409)

        vnf_package.onboarding_state = (
            fields.PackageOnboardingStateType.UPLOADING)

        try:
            vnf_package.save()
        except Exception as e:
            return self._make_problem_detail(
                'Internal Server Error', str(e), 500)

        try:
            (location, size, checksum, multihash,
            loc_meta) = glance_store.store_csar(context, id, body)
        except exceptions.UploadFailedToGlanceStore:
            with excutils.save_and_reraise_exception():
                vnf_package.onboarding_state = (
                    fields.PackageOnboardingStateType.CREATED)
                try:
                    vnf_package.save()
                except Exception as e:
                    return self._make_problem_detail(
                        'Internal Server Error', str(e), 500)

        vnf_package.algorithm = CONF.vnf_package.hashing_algorithm
        vnf_package.hash = multihash
        vnf_package.location_glance_store = location
        vnf_package.size = size
        try:
            vnf_package.save()
        except Exception as e:
            vnf_package.onboarding_state = (
                fields.PackageOnboardingStateType.CREATED)
            try:
                vnf_package.save()
            except Exception as e:
                return self._make_problem_detail(
                    'Internal Server Error', str(e), 500)

            return self._make_problem_detail(
                'Internal Server Error', str(e), 500)

        # process vnf_package
        self.rpc_api.upload_vnf_package_content(context, vnf_package)

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.expected_errors((http_client.BAD_REQUEST, http_client.FORBIDDEN,
                           http_client.NOT_FOUND, http_client.CONFLICT))
    @validation.schema(vnf_packages.upload_from_uri)
    def upload_vnf_package_from_uri(self, request, id, body):

        context = request.environ['tacker.context']
        context.can(vnf_package_policies.VNFPKGM % 'upload_from_uri')

        url = body['addressInformation']
        if not utils.is_valid_url(url):
            msg = _("Vnf package url '%s' is invalid") % url
            raise webob.exc.HTTPBadRequest(explanation=msg)

        vnf_package = self._get_vnf_package(id, request)

        if vnf_package.onboarding_state != \
                fields.PackageOnboardingStateType.CREATED:
            msg = _("VNF Package %(id)s onboarding state is not "
                    "%(onboarding)s")
            raise webob.exc.HTTPConflict(explanation=msg % {"id": id,
                    "onboarding": fields.PackageOnboardingStateType.CREATED})

        vnf_package.onboarding_state = (
            fields.PackageOnboardingStateType.UPLOADING)

        vnf_package.save()

        # process vnf_package
        self.rpc_api.upload_vnf_package_from_uri(context, vnf_package,
                                            body['addressInformation'],
                                            user_name=body.get('userName'),
                                            password=body.get('password'))

    @wsgi.response(http_client.OK)
    @wsgi.expected_errors((http_client.BAD_REQUEST, http_client.FORBIDDEN,
                           http_client.NOT_FOUND, http_client.CONFLICT))
    @validation.schema(vnf_packages.patch)
    def patch(self, request, id, body):
        context = request.environ['tacker.context']
        context.can(vnf_package_policies.VNFPKGM % 'patch')

        old_vnf_package = self._get_vnf_package(id, request)
        vnf_package = old_vnf_package.obj_clone()

        user_data = body.get('userDefinedData')
        operational_state = body.get('operationalState')

        if operational_state:
            if vnf_package.onboarding_state == \
                    fields.PackageOnboardingStateType.ONBOARDED:
                if vnf_package.operational_state == operational_state:
                    msg = _("VNF Package %(id)s is already in "
                            "%(operationState)s operational state") % {
                        "id": id,
                        "operationState": vnf_package.operational_state}
                    raise webob.exc.HTTPConflict(explanation=msg)
                else:
                    # update vnf_package operational state,
                    # if vnf_package Onboarding State is ONBOARDED
                    vnf_package.operational_state = operational_state
            else:
                if not user_data:
                    msg = _("Updating operational state is not allowed for VNF"
                            " Package %(id)s when onboarding state is not "
                            "%(onboarded)s")
                    raise webob.exc.HTTPBadRequest(
                        explanation=msg % {"id": id, "onboarded": fields.
                            PackageOnboardingStateType.ONBOARDED})
        # update user data
        if user_data:
            for key, value in list(user_data.items()):
                if vnf_package.user_data.get(key) == value:
                    del user_data[key]

            if not user_data:
                msg = _("The userDefinedData provided in update request is as"
                        " the existing userDefinedData of vnf package %(id)s."
                        " Nothing to update.")
                raise webob.exc.HTTPConflict(
                    explanation=msg % {"id": id})
            vnf_package.user_data = user_data

        vnf_package.save()

        return self._view_builder.patch(old_vnf_package, vnf_package)

    @wsgi.response(http_client.OK)
    @wsgi.expected_errors((http_client.BAD_REQUEST, http_client.FORBIDDEN,
                           http_client.NOT_FOUND, http_client.NOT_ACCEPTABLE,
                           http_client.CONFLICT,
                           http_client.INTERNAL_SERVER_ERROR))
    def get_vnf_package_vnfd(self, request, id):
        context = request.environ['tacker.context']
        context.can(vnf_package_policies.VNFPKGM % 'get_vnf_package_vnfd')

        valid_accept_headers = ['application/zip', 'text/plain']
        accept_headers = request.headers['Accept'].split(',')
        for header in accept_headers:
            if header not in valid_accept_headers:
                msg = _("Accept header %(accept)s is invalid, it should be one"
                        " of these values: %(valid_values)s")
                raise webob.exc.HTTPNotAcceptable(
                    explanation=msg % {"accept": header,
                                       "valid_values": ",".join(
                                           valid_accept_headers)})

        vnf_package = self._get_vnf_package(id, request)

        if vnf_package.onboarding_state != \
                fields.PackageOnboardingStateType.ONBOARDED:
            msg = _("VNF Package %(id)s state is not "
                    "%(onboarded)s")
            raise webob.exc.HTTPConflict(explanation=msg % {"id": id,
                    "onboarded": fields.PackageOnboardingStateType.ONBOARDED})

        try:
            vnfd_files_and_data = self.rpc_api.\
                get_vnf_package_vnfd(context, vnf_package)
        except exceptions.FailedToGetVnfdData as e:
            LOG.error(e.msg)
            raise webob.exc.HTTPInternalServerError(explanation=str(e.msg))

        if 'text/plain' in accept_headers:
            # Checking for yaml files only. This is required when there is
            # TOSCA.meta file along with single yaml file.
            # In such case we need to return single yaml file.
            file_list = list(vnfd_files_and_data.keys())
            yaml_files = [file for file in file_list if file.endswith(
                ('.yaml', '.yml'))]
            if len(yaml_files) == 1:
                request.response.headers['Content-Type'] = 'text/plain'
                return vnfd_files_and_data[yaml_files[0]]
            elif 'application/zip' in accept_headers:
                request.response.headers['Content-Type'] = 'application/zip'
                return self._create_vnfd_zip(vnfd_files_and_data)
            else:
                msg = _("VNFD is implemented as multiple yaml files,"
                        " Accept header should be 'application/zip'.")
                raise webob.exc.HTTPBadRequest(explanation=msg)
        else:
            request.response.headers['Content-Type'] = 'application/zip'
            return self._create_vnfd_zip(vnfd_files_and_data)

    @wsgi.response(http_client.OK)
    @wsgi.expected_errors((http_client.BAD_REQUEST, http_client.FORBIDDEN,
                           http_client.NOT_FOUND, http_client.CONFLICT,
                           http_client.REQUESTED_RANGE_NOT_SATISFIABLE))
    def fetch_vnf_package_artifacts(self, request, id, artifact_path):
        context = request.environ['tacker.context']
        # get policy
        context.can(vnf_package_policies.VNFPKGM % 'fetch_artifact')

        # get vnf_package
        if not uuidutils.is_uuid_like(id):
            msg = _("Can not find requested vnf package: %s") % id
            raise webob.exc.HTTPNotFound(explanation=msg)

        try:
            vnf_package = vnf_package_obj.VnfPackage.get_by_id(
                request.context, id,
                expected_attrs=["vnf_artifacts"])
        except exceptions.VnfPackageNotFound:
            msg = _("Can not find requested vnf package: %s") % id
            raise webob.exc.HTTPNotFound(explanation=msg)

        if vnf_package.onboarding_state != \
                fields.PackageOnboardingStateType.ONBOARDED:
            msg = _("VNF Package %(id)s state is not "
                    "%(onboarded)s")
            raise webob.exc.HTTPConflict(explanation=msg % {"id": id,
                    "onboarded": fields.PackageOnboardingStateType.ONBOARDED})

        offset, chunk_size = 0, None

        # get all artifact's path
        artifact_file_paths = []
        for item in vnf_package.vnf_artifacts:
            artifact_file_paths.append(item.artifact_path)

        if artifact_path in artifact_file_paths:
            # get file's size
            csar_path = self._get_csar_path(vnf_package)
            absolute_artifact_path = os.path.join(csar_path, artifact_path)
            if not os.path.isfile(absolute_artifact_path):
                msg = _(
                    "This type of path(url) '%s' is currently not supported") \
                    % artifact_path
                raise webob.exc.HTTPBadRequest(explanation=msg)
            artifact_size = os.path.getsize(absolute_artifact_path)
            range_val = self._get_range_from_request(request, artifact_size)
            # range_val exists
            if range_val:
                if isinstance(range_val, webob.byterange.Range):
                    # get the position of the last byte in the artifact file
                    response_end = artifact_size - 1
                    if range_val.start >= 0:
                        offset = range_val.start
                    else:
                        if abs(range_val.start) < artifact_size:
                            offset = artifact_size + range_val.start
                    if range_val.end is not None and \
                            range_val.end < artifact_size:
                        chunk_size = range_val.end - offset
                        response_end = range_val.end - 1
                    else:
                        chunk_size = artifact_size - offset
                request.response.status_int = 206
            # range_val does not exist, download the whole content of file
            else:
                offset = 0
                chunk_size = artifact_size

            # get file's mineType;
            mime_type = mimetypes.guess_type(artifact_path.split('/')[-1])[0]
            if mime_type:
                request.response.headers['Content-Type'] = mime_type
            else:
                request.response.headers['Content-Type'] = \
                    'application/octet-stream'
            try:
                artifact_data = self._download_vnf_artifact(
                    absolute_artifact_path, offset, chunk_size)
            except exceptions.FailedToGetVnfArtifact as e:
                LOG.error(e.msg)
                raise webob.exc.HTTPInternalServerError(
                    explanation=e.msg)
            request.response.text = artifact_data.decode('utf-8')
            if request.response.status_int == 206:
                request.response.headers['Content-Range'] = 'bytes %s-%s/%s' \
                                                            % (offset,
                                                               response_end,
                                                               artifact_size)
            else:
                chunk_size = artifact_size

            request.response.headers['Content-Length'] = chunk_size
            return request.response
        else:
            msg = _("Not Found Artifact File.")
            raise webob.exc.HTTPNotFound(explanation=msg)

    def _get_csar_path(self, vnf_package):
        csar_path = os.path.join(CONF.vnf_package.vnf_package_csar_path,
                                 vnf_package.id)

        if not os.path.isdir(csar_path):
            location = vnf_package.location_glance_store
            try:
                zip_path = glance_store.load_csar(vnf_package.id, location)
                csar_utils.extract_csar_zip_file(zip_path, csar_path)
            except (store_exceptions.GlanceStoreException) as e:
                exc_msg = encodeutils.exception_to_unicode(e)
                msg = (_("Exception raised from glance store can be "
                         "unrecoverable if it is not related to connection"
                         " error. Error: %s.") % exc_msg)
                raise exceptions.FailedToGetVnfArtifact(error=msg)
        return csar_path

    def _download_vnf_artifact(self, artifact_file_path, offset=0,
            chunk_size=None):
        try:
            with open(artifact_file_path, 'rb') as f:
                f.seek(offset, 1)
                vnf_artifact_data = f.read(chunk_size)
                return vnf_artifact_data
        except Exception as e:
            exc_msg = encodeutils.exception_to_unicode(e)
            msg = (_("Exception raised while reading artifact file"
                     " Error: %s.") % exc_msg)
            raise exceptions.FailedToGetVnfArtifact(error=msg)

    def _create_vnfd_zip(self, vnfd_files_and_data):
        buff = BytesIO()
        with ZipFile(buff, 'w', zipfile.ZIP_DEFLATED) as zip_archive:
            for file_path, file_data in vnfd_files_and_data.items():
                zip_archive.writestr(file_path, file_data)

        return buff.getvalue()

    def _make_problem_detail(self, title, detail, status):
        res = webob.Response(content_type='application/problem+json')
        problemDetails = {}
        problemDetails['title'] = title
        problemDetails['detail'] = detail
        problemDetails['status'] = status
        res.text = json.dumps(problemDetails)
        res.status_int = status
        return res


def create_resource():
    body_deserializers = {
        'application/zip': wsgi.ZipDeserializer()
    }

    deserializer = wsgi.RequestDeserializer(
        body_deserializers=body_deserializers)
    return wsgi.Resource(VnfPkgmController(), deserializer=deserializer)



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from tacker.tests import constants
from tacker.tests.functional import base


class VnfTestToscaFloatingIp(base.BaseTackerTest):

    def get_heat_stack_resource(self, stack_id, resource_name):
        resource_types = self.h_client.resources
        resource_details = resource_types.get(stack_id=stack_id,
                                              resource_name=resource_name)
        resource_dict = resource_details.to_dict()
        return resource_dict

    def connect_public_and_private_nw_with_router(self):
        public_nw = 'public'
        private_nw = 'net_mgmt'
        private_nw_subnet = 'subnet_mgmt'
        public_nw_id = None
        private_nw_id = None
        private_nw_subnet_id = None
        neutronclient = self.neutronclient()
        networks = neutronclient.list_networks()['networks']
        for nw in networks:
            if nw['name'] == public_nw:
                public_nw_id = nw['id']
            if nw['name'] == private_nw:
                private_nw_id = nw['id']
            if public_nw_id and private_nw_id:
                break
        self.assertIsNotNone(public_nw_id)
        self.assertIsNotNone(private_nw_id)
        subnets = neutronclient.list_subnets()['subnets']
        for subnet in subnets:
            if subnet['network_id'] == private_nw_id\
                    and subnet['name'] == private_nw_subnet:
                private_nw_subnet_id = subnet['id']
                break
        self.assertIsNotNone(private_nw_subnet_id)
        router_id = neutronclient.create_router(
            {'router': {'name': 'fip_test_router'}})['router']['id']
        self.assertIsNotNone(router_id)
        self.addCleanup(self.neutronclient().delete_router, router_id)
        rt_gw_id = neutronclient.add_gateway_router(
            router_id, {'network_id': public_nw_id})['router']['id']
        self.assertIsNotNone(rt_gw_id)
        self.addCleanup(self.neutronclient().remove_gateway_router,
            router_id)
        rt_int = neutronclient.add_interface_router(
            router_id, {'subnet_id': private_nw_subnet_id})['id']
        self.assertIsNotNone(rt_int)
        self.addCleanup(self.neutronclient().remove_interface_router,
            router_id, {'subnet_id': private_nw_subnet_id})

    def test_assign_floatingip_to_vdu(self):
        vnfd_file = 'sample_tosca_assign_floatingip_to_vdu.yaml'
        vnf_name = 'Assign Floating IP to VDU'
        self.connect_public_and_private_nw_with_router()
        vnfd_instance, vnf_instance, tosca_dict = self.vnfd_and_vnf_create(
            vnfd_file, vnf_name)
        vnf_id = vnf_instance['vnf']['id']
        self.addCleanup(self.wait_until_vnf_delete, vnf_id,
                        constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.addCleanup(self.client.delete_vnf, vnf_id)
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)
        vnf_show_out = self.client.show_vnf(vnf_id)['vnf']
        self.assertIsNotNone(vnf_show_out['mgmt_ip_address'])

        stack_id = vnf_show_out['instance_id']
        fip_res = self.get_heat_stack_resource(stack_id, 'FIP1')
        floating_ip_address = fip_res['attributes']['floating_ip_address']
        self.assertIsNotNone(floating_ip_address)
        fip_port_id = fip_res['attributes']['port_id']
        port_res = self.get_heat_stack_resource(stack_id, 'CP1')
        port_id = port_res['attributes']['id']
        self.assertEqual(fip_port_id, port_id)



# Copyright (C) 2022 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import shutil
import tempfile

from oslo_utils import uuidutils

from tacker.tests.functional.sol_v2 import paramgen
from tacker.tests.functional.sol_v2 import utils


zip_file_name = os.path.basename(os.path.abspath(".")) + '.zip'
tmp_dir = tempfile.mkdtemp()
vnfd_id = uuidutils.generate_uuid()
utils.make_zip(".", tmp_dir, vnfd_id)

shutil.copy(os.path.join(tmp_dir, zip_file_name), ".")
shutil.rmtree(tmp_dir)

subscription_req = paramgen.sub_create_min("http://127.0.0.1/")
create_req = paramgen.create_vnf_min(vnfd_id)
terminate_req = paramgen.terminate_vnf_min()
instantiate_req = paramgen.instantiate_vnf_min()

with open("subscription_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(subscription_req, indent=2))

with open("create_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(create_req, indent=2))

with open("terminate_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(terminate_req, indent=2))

with open("instantiate_req", "w", encoding='utf-8') as f:
    f.write(json.dumps(instantiate_req, indent=2))



# Copyright (c) 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pycodestyle
import re

from hacking import core

"""
Guidelines for writing new hacking checks

 - Use only for Tacker specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range N3xx. Find the current test with
   the highest allocated number and then pick the next value.
 - Keep the test method code in the source file ordered based
   on the N3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to
   tacker/tests/unit/test_hacking.py

"""

log_translation = re.compile(
    r"(.)*LOG\.(audit|error|info|warn|warning|critical|exception)\(\s*('|\")")


@core.flake8ext
def validate_log_translations(physical_line, logical_line, filename):
    # Translations are not required in the test directory
    if "tacker/tests" in filename:
        return
    if pycodestyle.noqa(physical_line):
        return
    msg = "N320: Log messages require translations!"
    if log_translation.match(logical_line):
        yield (0, msg)


@core.flake8ext
def no_log_warn(logical_line):
    """Disallow 'LOG.warn('

    Use LOG.warning() instead of Deprecated LOG.warn().
    https://docs.python.org/3/library/logging.html#logging.warning
    """

    msg = ("T301: LOG.warn is deprecated, please use LOG.warning!")
    if "LOG.warn(" in logical_line:
        yield (0, msg)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from copy import deepcopy
import datetime
import iso8601
import os
import shutil
import uuid
import webob
import yaml

from tacker.api.vnfpkgm.v1.router import VnfpkgmAPIRouter
from tacker import context
from tacker.objects import vnf_artifact as vnf_artifact_obj
from tacker.objects import vnf_deployment_flavour as vnf_deployment_flavour_obj
from tacker.objects import vnf_package as vnf_package_obj
from tacker.objects import vnf_package_vnfd as vnf_package_vnfd_obj
from tacker.objects import vnf_software_image as vnf_software_image_obj
from tacker.tests import constants
from tacker.tests import utils
from tacker.tests import uuidsentinel
from tacker import wsgi


VNFPACKAGE_RESPONSE = {
    '_links': {
        'packageContent': {
            'href': '/vnfpkgm/v1/vnf_packages/'
                    'f26f181d-7891-4720-b022-b074ec1733ef/package_content'},
            'self': {
                'href': '/vnfpkgm/v1/vnf_packages/'
                        'f26f181d-7891-4720-b022-b074ec1733ef'}
    },
    'checksum': {
        'algorithm': 'fake vnf package algorithm',
        'hash': 'fake vnf package hash'
    },
    'id': 'f26f181d-7891-4720-b022-b074ec1733ef',
    'onboardingState': 'ONBOARDED',
    'operationalState': 'ENABLED',
    'usageState': 'NOT_IN_USE',
    'vnfProductName': 'fake vnf product name',
    'vnfProvider': 'fake vnf provider',
    'vnfSoftwareVersion': 'fake vnf software version',
    'vnfdId': uuidsentinel.vnfd_id,
    'vnfdVersion': 'fake vnfd version',
    'userDefinedData': {'key1': 'value1', 'key2': 'value2'},
    'softwareImages': [{
        'checksum': {'algorithm': 'fake-algorithm',
                     'hash': 'fake software image hash'},
        'containerFormat': 'bare',
        'createdAt': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                       tzinfo=iso8601.UTC),
        'diskFormat': 'qcow2',
        'id': 'fake_software_image_id',
        'imagePath': 'fake image path',
        'minDisk': 1,
        'minRam': 0,
        'name': 'fake software image name',
        'provider': 'fake provider',
        'size': 1,
        'userMetadata': {'key3': 'value3', 'key4': 'value4'},
        'version': '11.22.33'
    }],
    'additionalArtifacts': [{
        'artifactPath': 'Scripts/install.sh',
        'metadata': {},
        'checksum': {
            'algorithm': 'SHA-256',
            'hash': '27bbdb25d8f4ed6d07d6f6581b86515e8b2f005'
                    '9b236ef7b6f50d6674b34f02a'
        }
    }],
}

VNFPACKAGE_INDEX_RESPONSE = [VNFPACKAGE_RESPONSE]


def index_response(remove_attrs=None, vnf_package_updates=None):
    # Returns VNFPACKAGE_INDEX_RESPONSE
    # parameter remove_attrs is a list of attribute names
    # to be removed before returning the response
    if not remove_attrs:
        return VNFPACKAGE_INDEX_RESPONSE
    vnf_package = deepcopy(VNFPACKAGE_RESPONSE)
    for attr in remove_attrs:
        vnf_package.pop(attr, None)
    if vnf_package_updates:
        vnf_package.update(vnf_package_updates)
    return [vnf_package]


def _fake_software_image(updates=None):
    software_image = {
        'id': uuidsentinel.software_image_id,
        'disk_format': 'qcow2',
        'min_ram': 0,
        'min_disk': 1,
        'container_format': 'bare',
        'provider': 'fake provider',
        'image_path': 'fake image path',
        'software_image_id': 'fake_software_image_id',
        'size': 1,
        'name': 'fake software image name',
        'hash': 'fake software image hash',
        'version': '11.22.33',
        'algorithm': 'fake-algorithm',
        'metadata': {'key3': 'value3', 'key4': 'value4'},
        'created_at': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                        tzinfo=iso8601.UTC),
    }
    if updates:
        software_image.update(updates)
    return software_image


def return_software_image(updates=None):
    software_image = _fake_software_image(updates)
    obj = vnf_software_image_obj.VnfSoftwareImage(**software_image)
    return obj


def _fake_deployment_flavour(updates=None):
    deployment_flavour = {
        'id': uuidsentinel.deployment_flavour_id,
        'package_uuid': 'f26f181d-7891-4720-b022-b074ec1733ef',
        'flavour_id': 'fake flavour id',
        'flavour_description': 'fake flavour description',
        'instantiation_levels': {"level1": 1, "level2": 2}
    }
    if updates:
        deployment_flavour.update(updates)
    return deployment_flavour


def _return_deployment_flavour(deployment_flavour_updates=None,
        software_image_updates=None):
    flavour = _fake_deployment_flavour(deployment_flavour_updates)
    obj = vnf_deployment_flavour_obj.VnfDeploymentFlavour(**flavour)

    software_image = return_software_image(software_image_updates)
    software_image_list = vnf_software_image_obj.VnfSoftwareImagesList()
    software_image_list.objects = [software_image]

    obj.software_images = software_image_list
    return obj


def _fake_vnfd(updates=None):
    vnfd = {
        'id': uuidsentinel.vnfd_unused_id,
        'package_uuid': 'f26f181d-7891-4720-b022-b074ec1733ef',
        'vnfd_id': uuidsentinel.vnfd_id,
        'vnf_provider': 'fake vnf provider',
        'vnf_product_name': 'fake vnf product name',
        'vnfd_version': 'fake vnfd version',
        'vnf_software_version': 'fake vnf software version'
    }
    if updates:
        vnfd.update(updates)
    return vnfd


def _return_vnfd(updates=None):
    vnfd = _fake_vnfd(updates)
    return vnf_package_vnfd_obj.VnfPackageVnfd(**vnfd)


def _fake_artifact(updates=None):
    artifact = {
        'id': uuidsentinel.vnf_artifact_id,
        'package_uuid': 'f26f181d-7891-4720-b022-b074ec1733ef',
        'artifact_path': 'Scripts/install.sh',
        'metadata': {},
        'algorithm': 'SHA-256',
        'hash': '27bbdb25d8f4ed6d07d6f6581b86515e8b2f0059b236ef'
                '7b6f50d6674b34f02a'}
    if updates:
        artifact.update(updates)
    return artifact


def _return_artifact(update=None):
    artifact = _fake_artifact(update)
    return vnf_artifact_obj.VnfPackageArtifactInfo(**artifact)


def _return_artifact_list(update=None):
    artifact_obj = _return_artifact(update)
    artifact_list = vnf_artifact_obj.VnfPackageArtifactInfoList()
    artifact_list.objects = [artifact_obj]
    return artifact_list


def fake_vnf_package(updates=None):
    vnf_package = {
        'id': constants.UUID,
        'hash': 'fake vnf package hash',
        'algorithm': 'fake vnf package algorithm',
        'location_glance_store': 'fake location',
        'onboarding_state': 'ONBOARDED',
        'operational_state': 'ENABLED',
        'tenant_id': uuidsentinel.tenant_id,
        'usage_state': 'NOT_IN_USE',
        'user_data': {'key1': 'value1', 'key2': 'value2'},
    }
    if updates:
        vnf_package.update(updates)
    return vnf_package


def return_vnfpkg_obj(
        vnf_package_updates=None,
        vnfd_updates=None,
        deployment_flavour_updates=None,
        software_image_updates=None,
        vnf_artifact_updates=None):
    vnf_package = fake_vnf_package(vnf_package_updates)
    obj = vnf_package_obj.VnfPackage(**vnf_package)
    obj.vnfd = _return_vnfd(vnfd_updates)

    deployment_flavour = _return_deployment_flavour(
        deployment_flavour_updates, software_image_updates)
    flavour_list = vnf_deployment_flavour_obj.VnfDeploymentFlavoursList()
    flavour_list.objects = [deployment_flavour]
    obj.vnf_deployment_flavours = flavour_list
    obj.vnf_artifacts = _return_artifact_list(vnf_artifact_updates)
    return obj


def return_vnf_package_list():
    vnf_package = return_vnfpkg_obj()
    return [vnf_package]


class InjectContext(wsgi.Middleware):
    """Add a 'tacker.context' to WSGI environ."""

    def __init__(self, context, *args, **kwargs):
        self.context = context
        super(InjectContext, self).__init__(*args, **kwargs)

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        req.environ['tacker.context'] = self.context
        return self.application


def wsgi_app_v1(fake_auth_context=None):
    inner_app_v1 = VnfpkgmAPIRouter()
    if fake_auth_context is not None:
        ctxt = fake_auth_context
    else:
        ctxt = context.ContextBase(uuidsentinel.user_id,
                                   uuidsentinel.project_id, is_admin=True)
    api_v1 = InjectContext(ctxt, inner_app_v1)
    return api_v1


def return_vnfd_data(csar_without_tosca_meta=False):
    csar_dir = ('sample_vnfpkg_no_meta_single_vnfd'
                if csar_without_tosca_meta else 'vnfpkgm1')
    unique_name = str(uuid.uuid4())
    csar_temp_dir = os.path.join('/tmp', unique_name)
    utils.copy_artifact_files(
        csar_temp_dir, csar_dir, csar_without_tosca_meta,
        read_vnfd_only=True)
    if csar_without_tosca_meta:
        file_names = ['vnfd_helloworld_single.yaml']
    else:
        file_names = ['TOSCA-Metadata/TOSCA.meta',
                      'Definitions/helloworld3_types.yaml',
                      'Definitions/helloworld3_df_simple.yaml',
                      'Definitions/helloworld3_top.vnfd.yaml',
                      'Definitions/etsi_nfv_sol001_common_types.yaml',
                      'Definitions/etsi_nfv_sol001_vnfd_types.yaml']
    file_path_and_data = {}
    for file_name in file_names:
        with open(os.path.join(csar_temp_dir, file_name)) as f:
            file_path_and_data.update({file_name: yaml.dump(
                yaml.safe_load(f))})

    shutil.rmtree(csar_temp_dir)
    return file_path_and_data



# Copyright (C) 2020 FUJITSU DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add db tables for add artifacts

Revision ID: e06fbdc90a32
Revises: d2e39e01d540
Create Date: 2020-09-17 02:52:41.435112

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = 'e06fbdc90a32'
down_revision = 'd2e39e01d540'

from alembic import op
import sqlalchemy as sa
from sqlalchemy import Boolean

from tacker.db import types


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'vnf_artifacts',
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('package_uuid', types.Uuid(length=36), nullable=False),
        sa.Column('artifact_path', sa.Text(), nullable=False),
        sa.Column('algorithm', sa.String(64), nullable=False),
        sa.Column('hash', sa.String(128), nullable=False),
        sa.Column('_metadata', sa.JSON(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['package_uuid'],
                                ['vnf_packages.id'], ),
        mysql_engine='InnoDB'
    )



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_serialization import jsonutils
from webob import Request

from tacker.alarm_receiver import AlarmReceiver
from tacker.tests.unit import base


class TestAlarmReceiver(base.TestCase):
    def setUp(self):
        '''url:

        http://tacker:9890/v1.0/vnfs/vnf-uuid/mon-policy-name/
        action-name/8ef785
        '''
        super(TestAlarmReceiver, self).setUp()
        self.alarmrc = AlarmReceiver(None)
        self.alarm_url = {
            '00_base': 'http://tacker:9890/v1.0',
            '01_url_base': '/vnfs/vnf-uuid/',
            '02_vnf_id': 'vnf-uuid',
            '03_monitoring_policy_name': 'mon-policy-name',
            '04_action_name': 'action-name',
            '05_key': '8ef785'
        }
        self.vnf_id = 'vnf-uuid'
        self.ordered_url = self._generate_alarm_url()

    def _generate_alarm_url(self):
        return 'http://tacker:9890/v1.0/vnfs/vnf-uuid/mon-policy-name/'\
               'action-name/8ef785'

    def test_handle_url(self):
        prefix_url, p, params = self.alarmrc.handle_url(self.ordered_url)
        self.assertEqual(self.alarm_url['01_url_base'], prefix_url)
        self.assertEqual(self.alarm_url['02_vnf_id'], p[3])
        self.assertEqual(self.alarm_url['03_monitoring_policy_name'], p[4])
        self.assertEqual(self.alarm_url['04_action_name'], p[5])
        self.assertEqual(self.alarm_url['05_key'], p[6])

    def test_handle_url_action_name(self):
        new_url = 'http://tacker:9890/v1.0/vnfs/vnf-uuid/mon-policy-name/'\
                  'respawn%25log/8ef785'
        prefix_url, p, params = self.alarmrc.handle_url(new_url)
        self.assertEqual(self.alarm_url['01_url_base'], prefix_url)
        self.assertEqual(self.alarm_url['02_vnf_id'], p[3])
        self.assertEqual(self.alarm_url['03_monitoring_policy_name'], p[4])
        self.assertEqual('respawn%log', p[5])
        self.assertEqual(self.alarm_url['05_key'], p[6])

    @mock.patch('tacker.vnfm.monitor_drivers.token.Token.create_token')
    def test_process_request(self, mock_create_token):
        mock_create_token.return_value = 'fake_token'
        req = Request.blank(self.ordered_url)
        req.method = 'POST'
        self.alarmrc.process_request(req)

        self.assertEqual(bytes('', 'utf-8'), req.body)
        self.assertEqual('fake_token', req.headers['X_AUTH_TOKEN'])
        self.assertIn(self.alarm_url['01_url_base'], req.environ['PATH_INFO'])
        self.assertIn('triggers', req.environ['PATH_INFO'])
        self.assertEqual('', req.environ['QUERY_STRING'])
        mock_create_token.assert_called_once_with()

    @mock.patch('tacker.vnfm.monitor_drivers.token.Token.create_token')
    def test_process_request_with_body(self, mock_create_token):
        req = Request.blank(self.ordered_url)
        req.method = 'POST'
        old_body = {'fake_key': 'fake_value'}
        req.body = jsonutils.dump_as_bytes(old_body)

        self.alarmrc.process_request(req)

        body_dict = jsonutils.loads(req.body)
        self.assertDictEqual(old_body,
                             body_dict['trigger']['params']['data'])
        self.assertEqual(self.alarm_url['05_key'],
                         body_dict['trigger']['params']['credential'])
        self.assertEqual(self.alarm_url['03_monitoring_policy_name'],
                         body_dict['trigger']['policy_name'])
        self.assertEqual(self.alarm_url['04_action_name'],
                         body_dict['trigger']['action_name'])



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os
import shutil
import time

from oslo_serialization import jsonutils
from oslo_utils import uuidutils
from sqlalchemy import desc
from sqlalchemy.orm import joinedload

from tacker.common import exceptions
from tacker import context
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker.objects import fields
from tacker.objects import vnf_lcm_op_occs
from tacker.tests.functional import base
from tacker.tests import utils
from tacker.vnfm.infra_drivers.kubernetes.kubernetes_driver import CONF

VNF_PACKAGE_UPLOAD_TIMEOUT = 300
RETRY_WAIT_TIME = 5

WAIT_TIMEOUT_ERR_MSG = ("Failed to %(action)s, process could not be completed"
                        " within %(timeout)s seconds")


class BaseVnfLcmKubernetesTest(base.BaseTackerTest):

    @classmethod
    def setUpClass(cls):
        super(BaseVnfLcmKubernetesTest, cls).setUpClass()
        cls.tacker_client = base.BaseTackerTest.tacker_http_client()
        cls.base_vnf_package_url = "/vnfpkgm/v1/vnf_packages"
        cls.base_vnf_instances_url = "/vnflcm/v1/vnf_instances"
        cls.base_vnf_lcm_op_occs_url = "/vnflcm/v1/vnf_lcm_op_occs"
        # [NOTE] culculate timeout from configuration. In addition to the sleep
        # time for each retry, add 1 second to timeout value for waiting
        # specified state as another processing time for each retry.
        timeout = (CONF.kubernetes_vim.stack_retries *
                   (CONF.kubernetes_vim.stack_retry_wait + 1))
        cls.lcm_timeout = {
            "instantiate": timeout,
            "terminate": timeout,
            "heal_sol002": timeout,
            "heal_sol003": timeout * 2,
            "scale": timeout,
            "modify": timeout
        }
        cls.vnf_package_ids = []

    @classmethod
    def tearDownClass(cls):
        # Update vnf package operational state to DISABLED and delete
        for package_id in cls.vnf_package_ids:
            cls._disable_and_delete_vnf_package(package_id)

        super(BaseVnfLcmKubernetesTest, cls).tearDownClass()

    def setUp(self):
        super(BaseVnfLcmKubernetesTest, self).setUp()
        self.context = context.get_admin_context()
        vim_list = self.client.list_vims()
        if not vim_list:
            self.skipTest("Vims are not configured")

        vim_name = 'vim-kubernetes'
        vim = self.get_vim(vim_list, vim_name)
        if not vim:
            self.skipTest(f"Kubernetes VIM '{vim_name}' is missing")
        self.vim_id = vim['id']

    def _create_and_upload_vnf_package_add_mgmt(
            self, tacker_client, csar_package_name,
            user_defined_data, mgmt_rela_path):
        # create vnf package
        body = jsonutils.dumps({"userDefinedData": user_defined_data})
        _, vnf_package = tacker_client.do_request(
            self.base_vnf_package_url, "POST", body=body)
        vnf_pkg_id = vnf_package['id']

        # cp MgmtDriver to package
        mgmt_abs_path = os.path.abspath(os.path.join(
            os.path.dirname(__file__), mgmt_rela_path))
        mgmt_package_abs_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__),
                         f"../../../etc/samples/etsi/nfv/"
                         f"{csar_package_name}/Scripts/"))
        os.mkdir(mgmt_package_abs_path)
        shutil.copy(mgmt_abs_path, mgmt_package_abs_path)

        # upload vnf package
        csar_package_path = ("../../../etc/samples/etsi/nfv/"
                             f"{csar_package_name}")
        file_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), csar_package_path))

        # Generating unique vnfd id. This is required when multiple workers
        # are running concurrently. The call below creates a new temporary
        # CSAR with unique vnfd id.
        file_path, _ = utils.create_csar_with_unique_vnfd_id(file_path)

        with open(file_path, 'rb') as file_object:
            tacker_client.do_request(
                f"{self.base_vnf_package_url}/{vnf_pkg_id}/package_content",
                "PUT", body=file_object, content_type='application/zip')

        # wait for onboard
        start_time = int(time.time())
        show_url = os.path.join(self.base_vnf_package_url, vnf_pkg_id)
        vnfd_id = None
        while True:
            _, body = tacker_client.do_request(show_url, "GET")
            if body['onboardingState'] == "ONBOARDED":
                vnfd_id = body['vnfdId']
                break

            if (int(time.time()) - start_time) > VNF_PACKAGE_UPLOAD_TIMEOUT:
                raise Exception(WAIT_TIMEOUT_ERR_MSG %
                    {"action": "onboard vnf package",
                     "timeout": VNF_PACKAGE_UPLOAD_TIMEOUT})

            time.sleep(RETRY_WAIT_TIME)

        # remove temporarily created CSAR file
        os.remove(file_path)
        shutil.rmtree(mgmt_package_abs_path)
        return vnf_package['id'], vnfd_id

    def _create_and_upload_vnf_package(self, tacker_client, csar_package_name,
                                       user_defined_data):
        # create vnf package
        body = jsonutils.dumps({"userDefinedData": user_defined_data})
        _, vnf_package = tacker_client.do_request(
            self.base_vnf_package_url, "POST", body=body)
        vnf_pkg_id = vnf_package['id']

        # upload vnf package
        csar_package_path = ("../../../etc/samples/etsi/nfv/"
                             f"{csar_package_name}")
        file_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                 csar_package_path))

        # Generating unique vnfd id. This is required when multiple workers
        # are running concurrently. The call below creates a new temporary
        # CSAR with unique vnfd id.
        file_path, _ = utils.create_csar_with_unique_vnfd_id(file_path)

        with open(file_path, 'rb') as file_object:
            tacker_client.do_request(
                f"{self.base_vnf_package_url}/{vnf_pkg_id}/package_content",
                "PUT", body=file_object, content_type='application/zip')

        # wait for onboard
        start_time = int(time.time())
        show_url = os.path.join(self.base_vnf_package_url, vnf_pkg_id)
        vnfd_id = None
        while True:
            _, body = tacker_client.do_request(show_url, "GET")
            if body['onboardingState'] == "ONBOARDED":
                vnfd_id = body['vnfdId']
                break

            if (int(time.time()) - start_time) > VNF_PACKAGE_UPLOAD_TIMEOUT:
                raise Exception(WAIT_TIMEOUT_ERR_MSG %
                    {"action": "onboard vnf package",
                     "timeout": VNF_PACKAGE_UPLOAD_TIMEOUT})

            time.sleep(RETRY_WAIT_TIME)

        # remove temporarily created CSAR file
        os.remove(file_path)
        return vnf_package['id'], vnfd_id

    @classmethod
    def _disable_and_delete_vnf_package(cls, package_id):
        # Update vnf package operational state to DISABLED
        update_req_body = jsonutils.dumps({
            "operationalState": "DISABLED"})
        cls.tacker_client.do_request(
            f'{cls.base_vnf_package_url}/{package_id}',
            "PATCH", content_type='application/json',
            body=update_req_body)

        # Delete vnf package
        url = f'{cls.base_vnf_package_url}/{package_id}'
        cls.tacker_client.do_request(url, "DELETE")

    @classmethod
    def _instantiate_vnf_instance_request(
            cls, flavour_id, vim_id=None, additional_param=None):
        request_body = {"flavourId": flavour_id}

        if vim_id:
            request_body["vimConnectionInfo"] = [
                {"id": uuidutils.generate_uuid(),
                 "vimId": vim_id,
                 "vimType": "kubernetes"}]

        if additional_param:
            request_body["additionalParams"] = additional_param

        return request_body

    def _create_vnf_instance(self, vnfd_id, vnf_instance_name=None,
                             vnf_instance_description=None):
        request_body = {'vnfdId': vnfd_id}
        if vnf_instance_name:
            request_body['vnfInstanceName'] = vnf_instance_name

        if vnf_instance_description:
            request_body['vnfInstanceDescription'] = vnf_instance_description

        resp, response_body = self.http_client.do_request(
            self.base_vnf_instances_url, "POST",
            body=jsonutils.dumps(request_body))
        self.assertEqual(201, resp.status_code)
        return resp, response_body

    def _delete_wait_vnf_instance(self, id):
        url = os.path.join(self.base_vnf_instances_url, id)
        start_time = int(time.time())
        while True:
            resp, _ = self.tacker_client.do_request(url, "DELETE")
            if 204 == resp.status_code:
                break

            if (int(time.time()) - start_time) > self.lcm_timeout['terminate']:
                raise Exception(WAIT_TIMEOUT_ERR_MSG %
                    {"action": "delete vnf instance",
                     "timeout": self.lcm_timeout['terminate']})
            time.sleep(RETRY_WAIT_TIME)

    def _show_vnf_instance(self, id):
        show_url = os.path.join(self.base_vnf_instances_url, id)
        _, vnf_instance = self.tacker_client.do_request(show_url, "GET")

        return vnf_instance

    def _vnf_instance_wait(
            self, id,
            instantiation_state=fields.VnfInstanceState.INSTANTIATED,
            timeout=None):
        show_url = os.path.join(self.base_vnf_instances_url, id)
        start_time = int(time.time())
        if timeout is None:
            timeout = self.lcm_timeout['instantiate']
        while True:
            _, body = self.tacker_client.do_request(show_url, "GET")
            if body['instantiationState'] == instantiation_state:
                break

            if (int(time.time()) - start_time) > timeout:
                raise Exception(WAIT_TIMEOUT_ERR_MSG %
                    {"action": "wait vnf instance", "timeout": timeout})

            time.sleep(RETRY_WAIT_TIME)

    def _instantiate_vnf_instance(self, id, request_body,
                                  wait_state="COMPLETED"):
        url = os.path.join(self.base_vnf_instances_url, id, "instantiate")
        resp, _ = self.http_client.do_request(
            url, "POST", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)
        if wait_state == "COMPLETED":
            self._vnf_instance_wait(id)
        # wait vnflcm_op_occs.operation_state become wait_state
        self._wait_vnflcm_op_occs(self.context, id,
                                  self.lcm_timeout['instantiate'], wait_state)

    def _create_and_instantiate_vnf_instance(self, vnfd_id, flavour_id,
                                             inst_name, inst_desc,
                                             additional_params):
        # create vnf instance
        _, vnf_instance = self._create_vnf_instance(
            vnfd_id, vnf_instance_name=inst_name,
            vnf_instance_description=inst_desc)
        self.assertEqual(
            'NOT_INSTANTIATED', vnf_instance['instantiationState'])

        # instantiate vnf instance
        additional_param = additional_params
        request_body = self._instantiate_vnf_instance_request(
            flavour_id, vim_id=self.vim_id, additional_param=additional_param)

        self._instantiate_vnf_instance(vnf_instance['id'], request_body)
        vnf_instance = self._show_vnf_instance(vnf_instance['id'])

        self.assertEqual(
            'INSTANTIATED', vnf_instance['instantiationState'])
        vnflcm_op_occ = self._get_vnflcm_op_occs_by_id(
            self.context, vnf_instance['id'])
        self.assertEqual('COMPLETED', vnflcm_op_occ.operation_state)
        self.assertEqual('INSTANTIATE', vnflcm_op_occ.operation)

        return vnf_instance

    def _modify_vnf_instance(self, vnf_instance_id, request_body):
        # modify vnf instance
        url = os.path.join(self.base_vnf_instances_url, vnf_instance_id)
        resp, _ = self.http_client.do_request(
            url, "PATCH", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)
        time.sleep(5)
        self._wait_vnflcm_op_occs(
            self.context, vnf_instance_id, self.lcm_timeout['modify'])
        vnflcm_op_occ = self._get_vnflcm_op_occs_by_id(
            self.context, vnf_instance_id)
        self.assertEqual('MODIFY_INFO', vnflcm_op_occ.operation)
        vnf_instance = self._show_vnf_instance(vnf_instance_id)
        vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)
        return vnf_instance, vnfc_rscs

    def _terminate_vnf_instance(self, id, request_body=None):
        if request_body is None:
            # Terminate vnf forcefully
            request_body = {
                "terminationType": fields.VnfInstanceTerminationType.FORCEFUL,
            }
        url = os.path.join(self.base_vnf_instances_url, id, "terminate")
        resp, _ = self.http_client.do_request(
            url, "POST", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)

        timeout = request_body.get('gracefulTerminationTimeout', None)
        start_time = int(time.time())

        self._vnf_instance_wait(
            id, instantiation_state=fields.VnfInstanceState.NOT_INSTANTIATED,
            timeout=self.lcm_timeout['terminate'])

        # If gracefulTerminationTimeout is set, check whether vnf
        # instantiation_state is set to NOT_INSTANTIATED after
        # gracefulTerminationTimeout seconds.
        if timeout and int(time.time()) - start_time < timeout:
            self.fail("Vnf is terminated before graceful termination "
                      "timeout period")

    def _delete_vnf_instance(self, id):
        self._delete_wait_vnf_instance(id)

        # verify vnf instance is deleted
        url = os.path.join(self.base_vnf_instances_url, id)
        resp, _ = self.http_client.do_request(url, "GET")
        self.assertEqual(404, resp.status_code)

    def _scale_vnf_instance(self, id, type, aspect_id,
                            number_of_steps=1):
        url = os.path.join(self.base_vnf_instances_url, id, "scale")
        # generate body
        request_body = {
            "type": type,
            "aspectId": aspect_id,
            "numberOfSteps": number_of_steps}
        resp, _ = self.http_client.do_request(
            url, "POST", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)

    def _heal_vnf_instance(self, id, vnfc_instance_id):
        url = os.path.join(self.base_vnf_instances_url, id, "heal")
        # generate body
        request_body = {
            "vnfcInstanceId": vnfc_instance_id}
        resp, _ = self.http_client.do_request(
            url, "POST", body=jsonutils.dumps(request_body))
        self.assertEqual(202, resp.status_code)

    @classmethod
    @db_api.context_manager.reader
    def _get_vnflcm_op_occs_by_id(cls, context, vnf_instance_id,
                                  columns_to_join=None):
        query = api.model_query(
            context, models.VnfLcmOpOccs,
            read_deleted="no", project_only=True).filter_by(
            vnf_instance_id=vnf_instance_id).order_by(
            desc("created_at"))

        if columns_to_join:
            for column in columns_to_join:
                query = query.options(joinedload(column))

        db_vnflcm_op_occ = query.first()

        if not db_vnflcm_op_occ:
            raise exceptions.VnfInstanceNotFound(id=vnf_instance_id)

        vnflcm_op_occ = vnf_lcm_op_occs.VnfLcmOpOcc.obj_from_db_obj(
            context, db_vnflcm_op_occ)
        return vnflcm_op_occ

    def _wait_vnflcm_op_occs(
            self, context, vnf_instance_id, timeout,
            operation_state='COMPLETED'):
        start_time = int(time.time())
        while True:
            vnflcm_op_occ = self._get_vnflcm_op_occs_by_id(
                context, vnf_instance_id)

            if vnflcm_op_occ.operation_state == operation_state:
                break

            if (int(time.time()) - start_time) > timeout:
                raise Exception("Timeout waiting for transition to"
                                f" {operation_state} state.")

            time.sleep(RETRY_WAIT_TIME)

    @classmethod
    def _get_vnfc_resource_info(cls, vnf_instance):
        inst_vnf_info = vnf_instance['instantiatedVnfInfo']
        vnfc_resource_info = inst_vnf_info['vnfcResourceInfo']
        return vnfc_resource_info

    def _get_scale_level_by_aspect_id(self, vnf_instance, aspect_id):
        scale_status = vnf_instance['instantiatedVnfInfo']['scaleStatus']
        self.assertTrue(len(scale_status) > 0)
        for status in scale_status:
            self.assertIsNotNone(status.get('aspectId'))
            self.assertIsNotNone(status.get('scaleLevel'))
            if status.get('aspectId') == aspect_id:
                scale_level = status.get('scaleLevel')
                break
        else:
            raise Exception(f"aspectId {aspect_id} is not found.")
        return scale_level

    def _test_scale(self, id, type, aspect_id, previous_level,
                    number_of_steps=1, error=False):
        # scale operation
        self._scale_vnf_instance(id, type, aspect_id, number_of_steps)
        wait_state = "COMPLETED"
        if error:
            expected_level = previous_level
            wait_state = "FAILED_TEMP"
        elif type == 'SCALE_OUT':
            expected_level = previous_level + number_of_steps
        else:
            expected_level = previous_level - number_of_steps
        # wait vnflcm_op_occs.operation_state become COMPLETE/FAILED_TEMP
        self._wait_vnflcm_op_occs(
            self.context, id, self.lcm_timeout['scale'], wait_state)
        # check scaleStatus after scale operation
        vnf_instance = self._show_vnf_instance(id)
        scale_level = self._get_scale_level_by_aspect_id(
            vnf_instance, aspect_id)
        self.assertEqual(scale_level, expected_level)

        return scale_level

    def _test_heal(self, vnf_instance, vnfc_instance_id):
        before_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)
        self._heal_vnf_instance(vnf_instance['id'], vnfc_instance_id)
        # wait vnflcm_op_occs.operation_state become COMPLETE
        if vnfc_instance_id:
            timeout = self.lcm_timeout['heal_sol002']
        else:
            timeout = self.lcm_timeout['heal_sol003']
        self._wait_vnflcm_op_occs(self.context, vnf_instance['id'], timeout)
        # check vnfcResourceInfo after heal operation
        vnf_instance = self._show_vnf_instance(vnf_instance['id'])
        after_vnfc_rscs = self._get_vnfc_resource_info(vnf_instance)
        self.assertEqual(len(before_vnfc_rscs), len(after_vnfc_rscs))
        return after_vnfc_rscs

    def _rollback_vnf_instance(self, vnf_lcm_op_occ_id):
        url = os.path.join(
            self.base_vnf_lcm_op_occs_url, vnf_lcm_op_occ_id, "rollback")
        # generate body
        resp, _ = self.http_client.do_request(url, "POST")
        self.assertEqual(202, resp.status_code)

    def _test_rollback_cnf_instantiate(self, id):
        # get vnflcm_op_occ id for rollback
        vnflcm_op_occ = self._get_vnflcm_op_occs_by_id(
            self.context, id)
        vnf_lcm_op_occ_id = vnflcm_op_occ.id

        # rollback operation
        self._rollback_vnf_instance(vnf_lcm_op_occ_id)
        # wait vnflcm_op_occs.operation_state become ROLLED_BACK
        self._wait_vnflcm_op_occs(self.context, id,
            self.lcm_timeout['terminate'], "ROLLED_BACK")

    def _test_rollback_cnf_scale(self, id, aspect_id, previous_level):
        # get vnflcm_op_occ id for rollback
        vnflcm_op_occ = self._get_vnflcm_op_occs_by_id(self.context, id)
        vnf_lcm_op_occ_id = vnflcm_op_occ.id

        # rollback operation
        self._rollback_vnf_instance(vnf_lcm_op_occ_id)
        # wait vnflcm_op_occs.operation_state become ROLLED_BACK
        self._wait_vnflcm_op_occs(self.context, id, self.lcm_timeout['scale'],
                                  "ROLLED_BACK")
        # check scaleStatus after scale operation
        vnf_instance = self._show_vnf_instance(id)
        expected_level = previous_level
        scale_level = self._get_scale_level_by_aspect_id(
            vnf_instance, aspect_id)
        self.assertEqual(scale_level, expected_level)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# flake8: noqa: E402

"""add_column_to_vnf_lcm_op_occs

Revision ID: 3adac34764da
Revises: 62d18199909e
Create Date: 2021-02-16 16:19:12.100380

"""

# revision identifiers, used by Alembic.
revision = '3adac34764da'
down_revision = '7186440a306b'

from alembic import op

import sqlalchemy as sa
from tacker.db import migration


def upgrade(active_plugins=None, options=None):
    op.add_column('vnf_lcm_op_occs',
        sa.Column('grant_id', sa.VARCHAR(length=36), nullable=True))
    op.add_column('vnf_lcm_op_occs',
        sa.Column('changed_ext_connectivity', sa.JSON(), nullable=True))



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""audit_support_events

Revision ID: 4ee19c8a6d0a
Revises: acf941e54075
Create Date: 2016-06-07 03:16:53.513392

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '4ee19c8a6d0a'
down_revision = '941b5a6fff9e'

from alembic import op
import sqlalchemy as sa

from tacker.db import types


def upgrade(active_plugins=None, options=None):
    op.create_table('events',
        sa.Column('id', sa.Integer, nullable=False, autoincrement=True),
        sa.Column('resource_id', types.Uuid, nullable=False),
        sa.Column('resource_state', sa.String(64), nullable=False),
        sa.Column('resource_type', sa.String(64), nullable=False),
        sa.Column('event_type', sa.String(64), nullable=False),
        sa.Column('timestamp', sa.DateTime, nullable=False),
        sa.Column('event_details', types.Json),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


CONF = cfg.CONF

OPTS = [
    cfg.StrOpt(
        'endpoint_url',
        default='http://localhost:9890/',
        help="endpoint_url"),
    cfg.IntOpt(
        'subscription_num',
        default=100,
        help="Number of subscriptions"),
    cfg.IntOpt(
        'retry_num',
        default=3,
        help="Number of retry"),
    cfg.IntOpt(
        'retry_wait',
        default=10,
        help="Retry interval (sec)"),
    cfg.IntOpt(
        'retry_timeout',
        default=10,
        help="Retry timeout (sec)"),
    cfg.BoolOpt(
        'test_callback_uri',
        default=True,
        help="Test callbackUri"),
    cfg.IntOpt(
        'operation_timeout',
        default=60,
        help="LCM operation timeout (sec)"),
    cfg.BoolOpt(
        'verify_notification_ssl',
        default=True,
        help="Verify the certificate to send notification by ssl"),
    cfg.IntOpt(
        'lcm_op_occ_num',
        default=100,
        help="Number of lcm_op_occs contained in 1 page"),
    cfg.IntOpt(
        'vnf_instance_num',
        default=100,
        help="Number of vnf_instances contained in 1 page"),
    cfg.IntOpt(
        'nextpage_expiration_time',
        default=3600,
        help="Expiration time (sec) for paging")]

vnf_lcm_group = cfg.OptGroup('vnf_lcm',
    title='vnf_lcm options',
    help="Vnflcm options group")


def register_opts(conf):
    conf.register_group(vnf_lcm_group)
    conf.register_opts(OPTS, group=vnf_lcm_group)


def list_opts():
    return {vnf_lcm_group: OPTS}



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from oslo_log import log as logging
from oslo_serialization import jsonutils
from tacker.common.utils import MemoryUnit
from tacker.tosca import utils as tosca_utils

"""Define util functions that can be used in UserData.

As for how to use the function (`create _ * _ dict`), dict can be obtained
as a return value by setting the required arguments and calling.
For detailed usage, check the docstring of each function.

NOTE: The functions defined here are limited to common and basic ones.
Please note that not all conversion logics are offered in Tacker,
different from Heat-Translator.
"""

LOG = logging.getLogger(__name__)

HOT_NOVA_SERVER = 'OS::Nova::Server'
HOT_NOVA_FLAVOR = 'OS::Nova::Flavor'
HOT_NEUTRON_PORT = 'OS::Neutron::Port'
AUTO_SCALING_GROUP = 'OS::Heat::AutoScalingGroup'
HOT_CINDER_VOLUME = 'OS::Cinder::Volume'
SUPPORTED_HOT_TYPE = [HOT_NOVA_SERVER, HOT_NOVA_FLAVOR,
                      HOT_NEUTRON_PORT, HOT_CINDER_VOLUME]
PORT_SERVER_TYPE = [HOT_NOVA_SERVER, HOT_NEUTRON_PORT, HOT_CINDER_VOLUME]


def create_initial_param_dict(base_hot_dict):
    """Create initial dict containing information about get_param resources.

    :param base_hot_dict: dict(Base HOT dict format)
    :return: dict('nfv', Initial HOT resource dict)

    NOTE: 'nfv' is a fixed value for 1st element.
          'VDU' and 'CP' are supported for 2nd element.
          3rd and 4th element are mandatory.
    """
    initial_param_dict = {
        'nfv': {
            'VDU': {
            },
            'CP': {
            }
        }
    }

    for hot_dict in base_hot_dict.values():
        param_nfv = hot_dict.get('parameters', {}).get('nfv')
        if not param_nfv:
            continue
        resources = hot_dict.get('resources', {})
        for resource_name, resource_val in resources.items():
            resource_type = resource_val.get('type')
            if resource_type in SUPPORTED_HOT_TYPE:
                resource_props = resource_val.get('properties', {})
                for prop_key, prop_val in resource_props.items():
                    if isinstance(prop_val, dict) and 'get_param' in prop_val:
                        param_list = prop_val['get_param']
                        if len(param_list) == 4:
                            resource_info = initial_param_dict.get(
                                param_list[0], {}).get(
                                    param_list[1], {})
                            if param_list[2] not in resource_info:
                                resource_info[param_list[2]] = {}
                if resource_type == HOT_NOVA_SERVER:
                    cinder_boot = resource_props.get('block_device_mapping_v2',
                                                     {})
                    if cinder_boot:
                        for boot in cinder_boot:
                            b_param = boot.get('image')
                            if b_param and \
                               isinstance(b_param, dict) and \
                               'get_param' in b_param:
                                param_list = b_param['get_param']
                                if len(param_list) == 4:
                                    resource_info = initial_param_dict.get(
                                        param_list[0], {}).get(
                                            param_list[1], {})
                                    if param_list[2] not in resource_info:
                                        resource_info[param_list[2]] = {}
            elif resource_type == AUTO_SCALING_GROUP:
                resource_nest = resource_val.get('properties').get('resource')
                resource_props = resource_nest.get('properties', {})
                for prop_key, prop_val in resource_props.items():
                    if isinstance(prop_val, dict) and 'get_param' in prop_val:
                        param_list = prop_val['get_param']
                        if len(param_list) == 4:
                            resource_info = initial_param_dict.get(
                                param_list[0], {}).get(
                                    param_list[1], {})
                            if param_list[2] not in resource_info:
                                resource_info[param_list[2]] = {}

    LOG.info('initial_param_dict: %s', initial_param_dict)
    return initial_param_dict


def create_initial_param_server_port_dict(base_hot_dict):
    """Create initial dict containing information about get_param resources.

    :param base_hot_dict: dict(Base HOT dict format)
    :return: dict('nfv', Initial HOT resource dict)

    NOTE: 'nfv' is a fixed value for 1st element.
          'VDU' and 'CP' are supported for 2nd element.
          3rd and 4th element are mandatory.
    """
    initial_param_dict = {
        'nfv': {
            'VDU': {
            },
            'CP': {
            }
        }
    }

    for hot_dict in base_hot_dict.values():
        LOG.debug("init hot_dict: %s", hot_dict)
        param_nfv = hot_dict.get('parameters', {}).get('nfv')
        if not param_nfv:
            continue
        resources = hot_dict.get('resources', {})
        for resource_name, resource_val in resources.items():
            resource_type = resource_val.get('type')
            if resource_type in PORT_SERVER_TYPE:
                resource_props = resource_val.get('properties', {})
                for prop_key, prop_val in resource_props.items():
                    if isinstance(prop_val, dict) and 'get_param' in prop_val:
                        param_list = prop_val['get_param']
                        if len(param_list) == 4:
                            resource_info = initial_param_dict.get(
                                param_list[0], {}).get(
                                    param_list[1], {})
                            if param_list[2] not in resource_info:
                                resource_info[param_list[2]] = {}
                if resource_type == HOT_NOVA_SERVER:
                    cinder_boot = resource_props.get('block_device_mapping_v2',
                                                     {})
                    if cinder_boot:
                        for boot in cinder_boot:
                            b_param = boot.get('image')
                            if b_param and \
                               isinstance(b_param, dict) and \
                               'get_param' in b_param:
                                param_list = b_param['get_param']
                                if len(param_list) == 4:
                                    resource_info = initial_param_dict.get(
                                        param_list[0], {}).get(
                                            param_list[1], {})
                                    if param_list[2] not in resource_info:
                                        resource_info[param_list[2]] = {}
            elif resource_type == AUTO_SCALING_GROUP:
                resource_nest = resource_val.get('properties').get('resource')
                resource_props = resource_nest.get('properties', {})
                for prop_key, prop_val in resource_props.items():
                    if isinstance(prop_val, dict) and 'get_param' in prop_val:
                        param_list = prop_val['get_param']
                        if len(param_list) == 4:
                            resource_info = initial_param_dict.get(
                                param_list[0], {}).get(
                                    param_list[1], {})
                            if param_list[2] not in resource_info:
                                resource_info[param_list[2]] = {}

    LOG.info('initial_param_dict: %s', initial_param_dict)
    return initial_param_dict


def create_final_param_dict(param_dict, vdu_flavor_dict,
                            vdu_image_dict, cpd_vl_dict):
    """Create final dict containing information about HOT input parameter.

    :param param_dict: dict('nfv', Initial HOT resource dict)
    :param vdu_flavor_dict: dict(VDU name, VDU flavor dict)
    :param vdu_iamge_dict: dict(VDU name, Glance-image uuid)
    :param cpd_vl_dict: dict(external CPD ID, Neutron-network uuid)
    :return: dict('nfv', Final HOT resource dict)
    """
    final_param_dict = copy.deepcopy(param_dict)

    vdus = final_param_dict.get('nfv', {}).get('VDU', {})
    for target_vdu in vdus:
        vdus[target_vdu]['flavor'] = vdu_flavor_dict.get(target_vdu)
        vdus[target_vdu]['image'] = vdu_image_dict.get(target_vdu)

    cps = final_param_dict.get('nfv', {}).get('CP', {})
    if cpd_vl_dict:
        for target_cp in cps:
            if 'network' in cpd_vl_dict.get(target_cp):
                cps[target_cp]['network'] = cpd_vl_dict.get(
                    target_cp).get('network')
            if 'fixed_ips' in cpd_vl_dict.get(target_cp):
                cps[target_cp]['fixed_ips'] = []
                for fixed_ip in cpd_vl_dict.get(target_cp).get("fixed_ips"):
                    cps[target_cp]['fixed_ips'].append(fixed_ip)

    LOG.info('final_param_dict: %s', final_param_dict)
    return final_param_dict


def create_vdu_flavor_dict(vnfd_dict):
    """Create a dict containing information about VDU's flavor.

    :param vnfd_dict: dict(VNFD dict format)
    :return: dict(VDU name, VDU flavor dict)
    """
    vdu_flavor_dict = {}
    node_templates = vnfd_dict.get(
        'topology_template', {}).get(
            'node_templates', {})

    for vdu_name, val in node_templates.items():
        vdu_flavor_props = val.get(
            'capabilities', {}).get(
                'virtual_compute', {}).get('properties', {})
        if vdu_flavor_props:
            flavor_dict = {}
            for key, val in vdu_flavor_props.items():
                if key == 'virtual_cpu':
                    flavor_dict['vcpus'] = val['num_virtual_cpu']
                elif key == 'virtual_memory':
                    # Convert to MiB
                    flavor_dict['ram'] = MemoryUnit.convert_unit_size_to_num(
                        val['virtual_mem_size'], 'MiB')
                elif key == 'virtual_local_storage':
                    # Convert to GiB
                    flavor_dict['disk'] = MemoryUnit.convert_unit_size_to_num(
                        val[0]['size_of_storage'], 'GiB')
            vdu_flavor_dict[vdu_name] = flavor_dict

    LOG.info('vdu_flavor_dict: %s', vdu_flavor_dict)
    return vdu_flavor_dict


def create_vdu_image_dict(grant_info):
    """Create a dict containing information about VDU's image.

    :param grant_info: dict(Grant information format)
    :return: dict(VDU name, Glance-image uuid)
    """
    vdu_image_dict = {}
    for vdu_name, resources in grant_info.items():
        for vnf_resource in resources:
            vdu_image_dict[vdu_name] = vnf_resource.resource_identifier

    LOG.info('vdu_image_dict: %s', vdu_image_dict)
    return vdu_image_dict


def create_cpd_vl_dict(base_hot_dict, inst_req_info):
    """Create a dict containing information about CPD and VL.

    :param base_hot_dict: dict(Base HOT dict format)
    :param inst_req_info: dict(Instantiation request information format)
    :return: dict(external CPD ID, Neutron-network uuid)
    """
    cpd_vl_dict = {}
    ext_vls = inst_req_info.ext_virtual_links
    if ext_vls is not None:
        for ext_vl in ext_vls:
            ext_cps = ext_vl.ext_cps
            vl_uuid = ext_vl.resource_id
            for ext_cp in ext_cps:
                for hot_dict in base_hot_dict.values():
                    cp_resource = hot_dict['resources'].get(
                        ext_cp.cpd_id)
                    if cp_resource is None:
                        continue
                    cpd_vl_dict[ext_cp.cpd_id] = vl_uuid
                    break

    LOG.info('cpd_vl_dict: %s', cpd_vl_dict)
    return cpd_vl_dict


def get_diff_base_hot_param_from_api(base_hot_dict, inst_req_info):
    """Compare base hot param from API param.

    :param base_hot_dict: dict(Base HOT dict format)
    :param inst_req_info: dict(Instantiation request information format)
    :return: dict(Parameters)
    """
    param_value = {}
    additional_param = inst_req_info.additional_params

    if additional_param is None:
        additional_param = {}
    input_attributes = base_hot_dict['heat_template'].get('parameters')

    for input_attr, value in input_attributes.items():
        if additional_param.get(input_attr):
            param_value.update({input_attr: additional_param.get(
                input_attr)})

    return param_value


def create_vdu_flavor_capability_name_dict(vnfd_dict):
    """Create a dict containing information about VDU's flavor.

    :param vnfd_dict: dict(VNFD dict format)
    :return: dict(VDU name, VDU Capability Name dict)
    """
    vdu_flavor_dict = {}
    node_templates = vnfd_dict.get(
        'topology_template', {}).get(
            'node_templates', {})

    for vdu_name, val in node_templates.items():
        vdu_flavor_props = val.get(
            'capabilities', {}).get(
                'virtual_compute', {}).get('properties', {})
        if vdu_flavor_props:
            for key, val in vdu_flavor_props.items():
                if key == 'requested_additional_capabilities':
                    capability_props = val.get('properties', {})
                    if 'requested_additional_capability_name'\
                       in capability_props.keys():
                        vdu_flavor_dict[vdu_name] = \
                            capability_props[
                                "requested_additional"
                                "_capability_name"]

    LOG.info('vdu_flavor_dict: %s', vdu_flavor_dict)
    return vdu_flavor_dict


def create_sw_image_dict(vnfd_dict):
    """Create a dict containing information about VDU's flavor.

    :param vnfd_dict: dict(VNFD dict format)
    :return: dict(VDU name, VDU SW Image data dict)
    """
    sw_image_data = {}
    node_templates = vnfd_dict.get(
        'topology_template', {}).get(
            'node_templates', {})

    for vdu_name, val in node_templates.items():
        sw_image_data_props = val.get(
            'properties', {}).get('sw_image_data', {})
        if sw_image_data_props:
            if 'name' in sw_image_data_props.keys():
                sw_image_data[vdu_name] = sw_image_data_props['name']

    LOG.info('sw_image_data: %s', sw_image_data)
    return sw_image_data


def create_network_dict(inst_req_info, param_dict):
    """Create a dict containing information about VDU's network.

    :param inst_req_info: dict(Instantiation request information format)
    :param param_dict: dict('nfv', Initial HOT resource dict)
    :return: dict(VDU name, VDU SW Image data dict)
    """
    cp_data = {}
    ext_vl_param = inst_req_info.ext_virtual_links
    cp_param = param_dict.get('nfv', {}).get('CP')

    for ext_vl in ext_vl_param:
        for ext_cp in ext_vl.ext_cps:
            if ext_cp.cpd_id in cp_param.keys():
                cp_data[ext_cp.cpd_id] = {}
                cp_data[ext_cp.cpd_id]["network"] = ext_vl.resource_id
                cp_data[ext_cp.cpd_id]["fixed_ips"] =\
                    _create_fixed_ips_list(ext_cp)

    LOG.info('cp_data: %s', cp_data)
    return cp_data


def _create_fixed_ips_list(ext_cp):
    """Create a list containing information about IP information.

    :param ext_cp: dict(ExtCp data of Instantiation request)
    :return: list(List structured of fixed_ips)
    """
    fixed_ips_lst = []
    for cp_config in ext_cp.cp_config:
        for cp_protocol_data in cp_config.cp_protocol_data:
            for ip_address in cp_protocol_data.ip_over_ethernet.ip_addresses:

                # TODO(w-juso):
                # Currently, I have not found a way to specify multiple
                # numDynamicAddresses to HOT, so I create a list of fixed_ips
                # with numDynamicAddress=1.
                # Needs to be considered in the future.
                fixed_ips = {}
                if ip_address.subnet_id:
                    fixed_ips['subnet'] = ip_address.subnet_id
                if ip_address.fixed_addresses:
                    for fixed_address in ip_address.fixed_addresses:
                        fixed_ips['ip_address'] = fixed_address

                if fixed_ips:
                    fixed_ips_lst.append(fixed_ips)

    return fixed_ips_lst


def _create_scale_group_dict(base_hot_dict, vnfd_dict, inst_req_info):

    base_hot = base_hot_dict['heat_template']
    LOG.debug("base_hot: %s", base_hot)

    scaling_group_dict = {}
    for name, rsc in base_hot.get('resources').items():
        if rsc['type'] == 'OS::Heat::AutoScalingGroup':
            key_name = name.replace('_group', '')
            scaling_group_dict[key_name] = name
    LOG.debug("scaling_group_dict: %s", scaling_group_dict)

    if scaling_group_dict:
        vnf = {'attributes': {'scaling_group_names':
            jsonutils.dump_as_bytes(scaling_group_dict)}}
        scale_group_dict = tosca_utils.get_scale_group(
            vnf, vnfd_dict, inst_req_info)
        LOG.debug("scale_group_dict: %s", scale_group_dict)
        return scale_group_dict
    else:
        LOG.debug("no scale_group_dict")
        return {}


def create_desired_capacity_dict(base_hot_dict, vnfd_dict, inst_req_info):
    """Create a dict containing information about desired capacity.

    :param base_hot_dict: dict(Base HOT dict format)
    :param vnfd_dict: dict(VNFD dict format)
    :param inst_req_info: dict(Instantiation request information format)
    :return: dict(Scaling aspect name, Desired capacity value)
    """
    scale_group_dict = _create_scale_group_dict(
        base_hot_dict, vnfd_dict, inst_req_info)

    param_dict = {}
    if scale_group_dict.get('scaleGroupDict'):
        for name, value in scale_group_dict['scaleGroupDict'].items():
            param_dict[name] = value['default']

    LOG.info("desired_capacity dict: %s", param_dict)
    return param_dict


def _calc_desired_capacity(inst_vnf_info, name, value):

    for scale_status in inst_vnf_info.scale_status:
        if scale_status.aspect_id == name:
            LOG.debug("scale_level of %s: %d",
                 name, scale_status.scale_level)
            increase = value['num'] * scale_status.scale_level
            desired_capacity = value['initialNum'] + increase
            LOG.debug("desired_capacity: %d", desired_capacity)
            return desired_capacity

    LOG.debug("scale_level of %s: None", name)
    return None


def get_desired_capacity_dict(base_hot_dict, vnfd_dict, inst_vnf_info):
    """Get a dict containing information about desired capacity.

    :param base_hot_dict: dict(Base HOT dict format)
    :param vnfd_dict: dict(VNFD dict format)
    :param inst_vnf_info: dict(Instantiated VNF Info dict format)
    :return: dict(Scaling aspect name, Desired capacity value)
    """
    scale_group_dict = _create_scale_group_dict(
        base_hot_dict, vnfd_dict, {})

    param_dict = {}
    if scale_group_dict.get('scaleGroupDict'):
        for name, value in scale_group_dict['scaleGroupDict'].items():
            desired_capacity = _calc_desired_capacity(
                inst_vnf_info, name, value)
            if desired_capacity is not None:
                param_dict[name] = desired_capacity

    LOG.info("desired_capacity dict: %s", param_dict)
    return param_dict
